@@ -1,10 +1,10 @@
 # -*- coding: utf-8 -*-
 
-"""Get triples from the Kinships data set."""
+"""Get triples from the Kinships dataset."""
 
 import os
 
-from ..base import PathDataSet
+from ..base import PathDataset
 
 __all__ = [
     'KINSHIPS_TRAIN_PATH',
@@ -20,13 +20,19 @@ KINSHIPS_TEST_PATH = os.path.join(HERE, 'test.txt')
 KINSHIPS_VALIDATE_PATH = os.path.join(HERE, 'valid.txt')
 
 
-class Kinships(PathDataSet):
-    """The Kinships data set."""
+class Kinships(PathDataset):
+    """The Kinships dataset."""
 
-    def __init__(self, **kwargs):
+    def __init__(self, create_inverse_triples: bool = False, **kwargs):
+        """Initialize the Kinships dataset.
+
+        :param create_inverse_triples: Should inverse triples be created? Defaults to false.
+        :param kwargs: keyword arguments passed to :class:`pykeen.datasets.base.PathDataset`.
+        """
         super().__init__(
             training_path=KINSHIPS_TRAIN_PATH,
             testing_path=KINSHIPS_TEST_PATH,
             validation_path=KINSHIPS_VALIDATE_PATH,
+            create_inverse_triples=create_inverse_triples,
             **kwargs,
         )
