@@ -9,15 +9,20 @@ from abc import ABC, abstractmethod
 from contextlib import contextmanager
 from dataclasses import dataclass
 from math import ceil
-from typing import Any, Collection, List, Mapping, Optional, Tuple, Union
+from typing import Any, Collection, Iterable, List, Mapping, Optional, Tuple, Union, cast
 
+import numpy as np
 import torch
-from dataclasses_json import dataclass_json
+from dataclasses_json import DataClassJsonMixin
+from tqdm.autonotebook import tqdm
 
-from ..models.base import Model
-from ..tqdmw import tqdm
+from ..models import Model
+from ..triples.utils import get_entities
 from ..typing import MappedTriples
-from ..utils import is_cuda_oom_error, is_cudnn_error, normalize_string, split_list_in_batches_iter
+from ..utils import (
+    is_cuda_oom_error, is_cudnn_error, is_nonzero_larger_than_maxint_error, normalize_string,
+    split_list_in_batches_iter,
+)
 
 __all__ = [
     'Evaluator',
@@ -38,9 +43,8 @@ def optional_context_manager(condition, context_manager):
         yield
 
 
-@dataclass_json
 @dataclass
-class MetricResults:
+class MetricResults(DataClassJsonMixin):
     """Results from computing metrics."""
 
     def get_metric(self, name: str) -> float:
@@ -64,13 +68,15 @@ class Evaluator(ABC):
         self,
         filtered: bool = False,
         requires_positive_mask: bool = False,
-        batch_size: int = None,
-        slice_size: int = None,
+        batch_size: Optional[int] = None,
+        slice_size: Optional[int] = None,
+        automatic_memory_optimization: bool = True,
     ):
         self.filtered = filtered
         self.requires_positive_mask = requires_positive_mask
         self.batch_size = batch_size
         self.slice_size = slice_size
+        self.automatic_memory_optimization = automatic_memory_optimization
 
     @classmethod
     def get_normalized_name(cls) -> str:
@@ -126,27 +132,39 @@ class Evaluator(ABC):
         slice_size: Optional[int] = None,
         device: Optional[torch.device] = None,
         use_tqdm: bool = True,
+        tqdm_kwargs: Optional[Mapping[str, str]] = None,
+        restrict_entities_to: Optional[torch.LongTensor] = None,
+        do_time_consuming_checks: bool = True,
     ) -> MetricResults:
         """Run :func:`pykeen.evaluation.evaluate` with this evaluator."""
         if mapped_triples is None:
             mapped_triples = model.triples_factory.mapped_triples
 
-        if batch_size is None and model.automatic_memory_optimization:
-            batch_size, slice_size = self.batch_and_slice(
-                model=model,
-                mapped_triples=mapped_triples,
-                batch_size=batch_size,
-                device=device,
-                use_tqdm=False,
-            )
-            # The batch_size and slice_size should be accessible to outside objects for re-use, e.g. early stoppers.
-            self.batch_size = batch_size
-            self.slice_size = slice_size
+        if batch_size is None and self.automatic_memory_optimization:
+            # Using automatic memory optimization on CPU may result in undocumented crashes due to OS' OOM killer.
+            if model.device.type == 'cpu':
+                logger.info(
+                    "Currently automatic memory optimization only supports GPUs, but you're using a CPU. "
+                    "Therefore, the batch_size will be set to the default value.",
+                )
+            else:
+                batch_size, slice_size = self.batch_and_slice(
+                    model=model,
+                    mapped_triples=mapped_triples,
+                    batch_size=batch_size,
+                    device=device,
+                    use_tqdm=False,
+                    restrict_entities_to=restrict_entities_to,
+                    do_time_consuming_checks=do_time_consuming_checks,
+                )
+                # The batch_size and slice_size should be accessible to outside objects for re-use, e.g. early stoppers.
+                self.batch_size = batch_size
+                self.slice_size = slice_size
 
-            # Clear the ranks from the current evaluator
-            self.finalize()
+                # Clear the ranks from the current evaluator
+                self.finalize()
 
-        return evaluate(
+        rv = evaluate(
             model=model,
             mapped_triples=mapped_triples,
             evaluators=self,
@@ -155,7 +173,12 @@ class Evaluator(ABC):
             device=device,
             squeeze=True,
             use_tqdm=use_tqdm,
+            tqdm_kwargs=tqdm_kwargs,
+            restrict_entities_to=restrict_entities_to,
+            do_time_consuming_checks=do_time_consuming_checks,
         )
+        # Since squeeze is true, we can expect that evaluate returns a MetricResult, but we need to tell MyPy that
+        return cast(MetricResults, rv)
 
     def batch_and_slice(
         self,
@@ -164,6 +187,8 @@ class Evaluator(ABC):
         batch_size: Optional[int] = None,
         device: Optional[torch.device] = None,
         use_tqdm: bool = False,
+        restrict_entities_to: Optional[torch.LongTensor] = None,
+        do_time_consuming_checks: bool = True,
     ) -> Tuple[int, Optional[int]]:
         """Find the maximum possible batch_size and slice_size for evaluation with the current setting.
 
@@ -185,6 +210,8 @@ class Evaluator(ABC):
             The device on which the evaluation shall be run. If None is given, use the model's device.
         :param use_tqdm:
             Should a progress bar be displayed?
+        :param restrict_entities_to:
+            Whether to restrict the evaluation to certain entities of interest.
 
         :return:
             Maximum possible batch size and, if necessary, the slice_size, which defaults to None.
@@ -199,6 +226,8 @@ class Evaluator(ABC):
             mapped_triples=mapped_triples,
             device=device,
             use_tqdm=use_tqdm,
+            restrict_entities_to=restrict_entities_to,
+            do_time_consuming_checks=do_time_consuming_checks,
         )
 
         if evaluated_once:  # slice_size = None
@@ -214,6 +243,8 @@ class Evaluator(ABC):
             mapped_triples=mapped_triples,
             device=device,
             use_tqdm=use_tqdm,
+            restrict_entities_to=restrict_entities_to,
+            do_time_consuming_checks=False,
         )
         if not evaluated_once:
             raise MemoryError("The current model can't be trained on this hardware with these parameters.")
@@ -223,11 +254,13 @@ class Evaluator(ABC):
     def _param_size_search(
         self,
         key: str,
-        start_value: int,
+        start_value: Optional[int],
         model: Model,
         mapped_triples: MappedTriples,
         device: Optional[torch.device] = None,
         use_tqdm: bool = False,
+        restrict_entities_to: Optional[torch.LongTensor] = None,
+        do_time_consuming_checks: bool = True,
     ) -> Tuple[int, bool]:
         values_dict = {}
         maximum_triples = mapped_triples.shape[0]
@@ -239,11 +272,14 @@ class Evaluator(ABC):
             values_dict[key] = start_value
             values_dict['slice_size'] = None
         elif key == 'slice_size':
+            if start_value is None:
+                start_value = ceil(model.num_entities / 2)
             self._check_slicing_availability(model, batch_size=1)
             values_dict[key] = start_value
             values_dict['batch_size'] = 1
         else:
             raise AttributeError(f'The parameter {key} is unknown.')
+
         reached_max = False
         evaluated_once = False
         logger.info(f'Starting {key} search for evaluation now...')
@@ -254,7 +290,6 @@ class Evaluator(ABC):
                 gc.collect()
                 torch.cuda.empty_cache()
                 evaluate(
-                    **values_dict,
                     model=model,
                     mapped_triples=mapped_triples,
                     evaluators=self,
@@ -262,6 +297,10 @@ class Evaluator(ABC):
                     device=device,
                     squeeze=True,
                     use_tqdm=use_tqdm,
+                    restrict_entities_to=restrict_entities_to,
+                    do_time_consuming_checks=do_time_consuming_checks,
+                    batch_size=values_dict.get('batch_size'),
+                    slice_size=values_dict.get('slice_size'),
                 )
                 evaluated_once = True
             except RuntimeError as runtime_error:
@@ -272,15 +311,20 @@ class Evaluator(ABC):
                 # The cache of the previous run has to be freed to allow accurate memory availability estimates
                 gc.collect()
                 torch.cuda.empty_cache()
-                if not is_cudnn_error(runtime_error) and not is_cuda_oom_error(runtime_error):
+                if (
+                    not is_cudnn_error(runtime_error)
+                    and not is_cuda_oom_error(runtime_error)
+                    and not is_nonzero_larger_than_maxint_error(runtime_error)
+                ):
                     raise runtime_error
                 if values_dict[key] == 1:
                     logger.debug(
-                        f"Even {key} {values_dict[key]} does not fit into your memory with these parameters."
+                        f"Even {key} {values_dict[key]} does not fit into your memory with these parameters.",
                     )
                     break
 
-                values_dict[key] //= 2
+                #  values_dict[key] will always be an int at this point
+                values_dict[key] //= 2  # type: ignore
                 reached_max = True
                 if evaluated_once:
                     logger.info(f'Concluded {key} search with batch_size={values_dict[key]}.')
@@ -292,12 +336,12 @@ class Evaluator(ABC):
                 gc.collect()
                 torch.cuda.empty_cache()
                 if not reached_max and values_dict['batch_size'] < maximum_triples:
-                    values_dict[key] *= 2
+                    values_dict[key] *= 2  # type: ignore
                 else:
                     logger.info(f'Concluded {key} search with batch_size={values_dict[key]}.')
                     break
 
-        return values_dict[key], evaluated_once
+        return cast(Tuple[int, bool], (values_dict[key], evaluated_once))
 
     @staticmethod
     def _check_slicing_availability(model: Model, batch_size: int) -> None:
@@ -357,7 +401,7 @@ def create_sparse_positive_filter_(
     entities = hrt_batch[:, other_col:other_col + 1]
 
     entity_filter_test = (all_pos_triples[:, other_col:other_col + 1]).view(1, -1) == entities
-    filter_batch = (entity_filter_test & relation_filter).nonzero()
+    filter_batch = (entity_filter_test & relation_filter).nonzero(as_tuple=False)
     filter_batch[:, 1] = all_pos_triples[:, filter_col:filter_col + 1].view(1, -1)[:, filter_batch[:, 1]]
 
     return filter_batch, relation_filter
@@ -422,6 +466,9 @@ def evaluate(
     device: Optional[torch.device] = None,
     squeeze: bool = True,
     use_tqdm: bool = True,
+    tqdm_kwargs: Optional[Mapping[str, str]] = None,
+    restrict_entities_to: Optional[torch.LongTensor] = None,
+    do_time_consuming_checks: bool = True,
 ) -> Union[MetricResults, List[MetricResults]]:
     """Evaluate metrics for model on mapped triples.
 
@@ -434,7 +481,8 @@ def evaluate(
     :param model:
         The model to evaluate.
     :param mapped_triples:
-        The triples on which to evaluate.
+        The triples on which to evaluate. The mapped triples should never contain inverse triples - these are created by
+        the model class on the fly.
     :param evaluators:
         An evaluator or a list of evaluators working on batches of triples and corresponding scores.
     :param only_size_probing:
@@ -449,12 +497,31 @@ def evaluate(
         Return a single instance of :class:`MetricResults` if only one evaluator was given.
     :param use_tqdm:
         Should a progress bar be displayed?
+    :param restrict_entities_to:
+        Optionally restrict the evaluation to the given entity IDs. This may be useful if one is only interested in a
+        part of the entities, e.g. due to type constraints, but wants to train on all available data. For ranking the
+        entities, we still compute all scores for all possible replacement entities to avoid irregular access patterns
+        which might decrease performance, but the scores with afterwards be filtered to only keep those of interest.
+        If provided, we assume that the triples are already filtered, such that it only contains the entities of
+        interest.
+    :param do_time_consuming_checks:
+        Whether to perform some time consuming checks on the provided arguments. Currently, this encompasses:
+        - If restrict_entities_to is not None, check whether the triples have been filtered.
+        Disabling this option can accelerate the method.
     """
     if isinstance(evaluators, Evaluator):  # upgrade a single evaluator to a list
         evaluators = [evaluators]
 
     start = timeit.default_timer()
 
+    # verify that the triples have been filtered
+    if restrict_entities_to is not None and do_time_consuming_checks:
+        present_entity_ids = get_entities(triples=mapped_triples)
+        unwanted = present_entity_ids.difference(restrict_entities_to.tolist())
+        if len(unwanted) > 0:
+            raise ValueError(f'mapped_triples contains IDs of entities which are not contained in restrict_entities_to:'
+                             f'{unwanted}. This will invalidate the evaluation results.')
+
     # Send to device
     if device is not None:
         model = model.to(device)
@@ -486,8 +553,10 @@ def evaluate(
 
     # Prepare batches
     if batch_size is None:
-        batch_size = 1
-    batches = split_list_in_batches_iter(input_list=mapped_triples, batch_size=batch_size)
+        # This should be a reasonable default size that works on most setups while being faster than batch_size=1
+        batch_size = 32
+        logger.info(f"No evaluation batch_size provided. Setting batch_size to '{batch_size}'.")
+    batches = cast(Iterable[np.ndarray], split_list_in_batches_iter(input_list=mapped_triples, batch_size=batch_size))
 
     # Show progressbar
     num_triples = mapped_triples.shape[0]
@@ -496,136 +565,36 @@ def evaluate(
     evaluated_once = False
 
     # Disable gradient tracking
-    with optional_context_manager(
-        use_tqdm,
-        tqdm(
-            desc=f'Evaluating on {model.device}',
-            total=num_triples,
-            unit='triple',
-            unit_scale=True,
-            # Choosing no progress bar (use_tqdm=False) would still show the initial progress bar without disable=True
-            disable=not use_tqdm,
-        ),
-    ) as progress_bar, torch.no_grad():
+    _tqdm_kwargs = dict(
+        desc=f'Evaluating on {model.device}',
+        total=num_triples,
+        unit='triple',
+        unit_scale=True,
+        # Choosing no progress bar (use_tqdm=False) would still show the initial progress bar without disable=True
+        disable=not use_tqdm,
+    )
+    if tqdm_kwargs:
+        _tqdm_kwargs.update(tqdm_kwargs)
+    with optional_context_manager(use_tqdm, tqdm(**_tqdm_kwargs)) as progress_bar, torch.no_grad():
         # batch-wise processing
         for batch in batches:
             batch_size = batch.shape[0]
-
-            # Predict tail scores once
-            scores_of_corrupted_tails_batch = model.predict_scores_all_tails(batch[:, 0:2], slice_size=slice_size)
-            scores_of_true_tails_batch = scores_of_corrupted_tails_batch[
-                torch.arange(0, batch.shape[0]),
-                batch[:, 2],
-            ]
-
-            # Create positive filter for all corrupted tails
-            if filtering_necessary or positive_masks_required:
-                assert all_pos_triples is not None
-                positive_filter_tails, relation_filter = create_sparse_positive_filter_(
-                    hrt_batch=batch,
-                    all_pos_triples=all_pos_triples,
-                    relation_filter=None,
-                    filter_col=2,
-                )
-
-            # Create a positive mask with the size of the scores from the positive tails filter
-            if positive_masks_required:
-                positive_mask_tails = create_dense_positive_mask_(
-                    zero_tensor=torch.zeros_like(scores_of_corrupted_tails_batch),
-                    filter_batch=positive_filter_tails,
-                )
-            else:
-                positive_mask_tails = None
-
-            # Evaluate metrics on these *unfiltered* tail scores
-            for unfiltered_evaluator in unfiltered_evaluators:
-                unfiltered_evaluator.process_tail_scores_(
-                    hrt_batch=batch,
-                    true_scores=scores_of_true_tails_batch[:, None],
-                    scores=scores_of_corrupted_tails_batch,
-                    dense_positive_mask=positive_mask_tails,
-                )
-
-            # Filter
-            if filtering_necessary:
-                filtered_scores_of_corrupted_tails_batch = filter_scores_(
-                    scores=scores_of_corrupted_tails_batch,
-                    filter_batch=positive_filter_tails,
-                )
-
-                # The scores for the true triples have to be rewritten to the scores tensor
-                scores_of_corrupted_tails_batch[
-                    torch.arange(0, batch.shape[0]),
-                    batch[:, 2],
-                ] = scores_of_true_tails_batch
-
-                # Evaluate metrics on these *filtered* tail scores
-                for filtered_evaluator in filtered_evaluators:
-                    filtered_evaluator.process_tail_scores_(
-                        hrt_batch=batch,
-                        true_scores=scores_of_true_tails_batch[:, None],
-                        scores=filtered_scores_of_corrupted_tails_batch,
-                    )
-
-            # Predict head scores once
-            scores_of_corrupted_heads_batch = model.predict_scores_all_heads(batch[:, 1:3], slice_size=slice_size)
-            scores_of_true_heads_batch = scores_of_corrupted_heads_batch[
-                torch.arange(0, batch.shape[0]),
-                batch[:, 0],
-            ]
-
-            # Create positive filter for all corrupted heads
-            if filtering_necessary or positive_masks_required:
-                assert all_pos_triples is not None
-                assert relation_filter is not None
-                positive_filter_heads, _ = create_sparse_positive_filter_(
-                    hrt_batch=batch,
+            relation_filter = None
+            for column in (0, 2):
+                relation_filter = _evaluate_batch(
+                    batch=batch,
+                    model=model,
+                    column=column,
+                    filtered_evaluators=filtered_evaluators,
+                    unfiltered_evaluators=unfiltered_evaluators,
+                    slice_size=slice_size,
                     all_pos_triples=all_pos_triples,
                     relation_filter=relation_filter,
-                    filter_col=0,
+                    restrict_entities_to=restrict_entities_to,
+                    positive_masks_required=positive_masks_required,
+                    filtering_necessary=filtering_necessary,
                 )
 
-            # Create a positive mask with the size of the scores from the positive heads filter
-            if positive_masks_required:
-                positive_mask_heads = create_dense_positive_mask_(
-                    zero_tensor=torch.zeros_like(scores_of_corrupted_heads_batch),
-                    filter_batch=positive_filter_heads,
-                )
-            else:
-                positive_mask_heads = None
-
-            # Evaluate metrics on these head scores
-            for evaluator in unfiltered_evaluators:
-                evaluator.process_head_scores_(
-                    hrt_batch=batch,
-                    true_scores=scores_of_true_heads_batch[:, None],
-                    scores=scores_of_corrupted_heads_batch,
-                    dense_positive_mask=positive_mask_heads,
-                )
-
-            # Filter
-            if filtering_necessary:
-                assert all_pos_triples is not None
-                assert relation_filter is not None
-                filtered_scores_of_corrupted_heads_batch = filter_scores_(
-                    scores=scores_of_corrupted_heads_batch,
-                    filter_batch=positive_filter_heads
-                )
-
-                # The scores for the true triples have to be rewritten to the scores tensor
-                scores_of_corrupted_heads_batch[
-                    torch.arange(0, batch.shape[0]),
-                    batch[:, 0],
-                ] = scores_of_true_heads_batch
-
-                # Evaluate metrics on these *filtered* tail scores
-                for filtered_evaluator in filtered_evaluators:
-                    filtered_evaluator.process_head_scores_(
-                        hrt_batch=batch,
-                        true_scores=scores_of_true_heads_batch[:, None],
-                        scores=filtered_scores_of_corrupted_heads_batch,
-                    )
-
             # If we only probe sizes we do not need more than one batch
             if only_size_probing and evaluated_once:
                 break
@@ -648,3 +617,136 @@ def evaluate(
         return results[0]
 
     return results
+
+
+def _evaluate_batch(
+    batch: MappedTriples,
+    model: Model,
+    column: int,
+    filtered_evaluators: Collection[Evaluator],
+    unfiltered_evaluators: Collection[Evaluator],
+    slice_size: Optional[int],
+    all_pos_triples: Optional[MappedTriples],
+    relation_filter: Optional[torch.BoolTensor],
+    restrict_entities_to: Optional[torch.LongTensor],
+    positive_masks_required: bool,
+    filtering_necessary: bool,
+) -> torch.BoolTensor:
+    """
+    Evaluate batch for all head predictions(column=0), or all tail predictions (column=2).
+
+    :param batch: shape: (batch_size, 3)
+        The batch of currently evaluated triples.
+    :param model:
+        The model to evaluate.
+    :param column:
+        The column which to evaluate. Either 0 for head prediction, or 2 for tail prediction.
+    :param filtered_evaluators:
+        The evaluators which work on filtered scores.
+    :param unfiltered_evaluators:
+        The evaluators which work on unfiltered scores.
+    :param slice_size:
+        An optional slice size for computing the scores.
+    :param all_pos_triples:
+        All positive triples (required if filtering is necessary).
+    :param relation_filter:
+        The relation filter. Can be re-used.
+    :param restrict_entities_to:
+        Restriction to evaluate only for these entities.
+    :param positive_masks_required:
+        Whether dense positive masks are required (by any unfiltered evaluator).
+    :param filtering_necessary:
+        Whether filtering is necessary.
+
+    :return:
+        The relation filter, which can be re-used for the same batch.
+    """
+    if column not in {0, 2}:
+        raise ValueError(f'column must be either 0 or 2, but is column={column}')
+
+    # Predict scores once
+    if column == 2:  # tail scores
+        batch_scores_of_corrupted = model.predict_t(batch[:, 0:2], slice_size=slice_size)
+    else:
+        batch_scores_of_corrupted = model.predict_h(batch[:, 1:3], slice_size=slice_size)
+
+    # Select scores of true
+    batch_scores_of_true = batch_scores_of_corrupted[
+        torch.arange(0, batch.shape[0]),
+        batch[:, column],
+    ]
+
+    # Create positive filter for all corrupted
+    if filtering_necessary or positive_masks_required:
+        # Needs all positive triples
+        if all_pos_triples is None:
+            raise ValueError('If filtering_necessary of positive_masks_required is True, all_pos_triples has to be '
+                             'provided, but is None.')
+
+        # Create filter
+        positive_filter, relation_filter = create_sparse_positive_filter_(
+            hrt_batch=batch,
+            all_pos_triples=all_pos_triples,
+            relation_filter=relation_filter,
+            filter_col=column,
+        )
+
+    # Create a positive mask with the size of the scores from the positive filter
+    if positive_masks_required:
+        positive_mask = create_dense_positive_mask_(
+            zero_tensor=torch.zeros_like(batch_scores_of_corrupted),
+            filter_batch=positive_filter,
+        )
+    else:
+        positive_mask = None
+
+    # Restrict to entities of interest
+    if restrict_entities_to is not None:
+        batch_scores_of_corrupted_ = batch_scores_of_corrupted[:, restrict_entities_to]
+        positive_mask = positive_mask[:, restrict_entities_to]
+    else:
+        batch_scores_of_corrupted_ = batch_scores_of_corrupted
+
+    # Evaluate metrics on these *unfiltered* scores
+    for unfiltered_evaluator in unfiltered_evaluators:
+        if column == 2:  # tail scores
+            process = unfiltered_evaluator.process_tail_scores_
+        else:
+            process = unfiltered_evaluator.process_head_scores_
+        process(
+            hrt_batch=batch,
+            true_scores=batch_scores_of_true[:, None],
+            scores=batch_scores_of_corrupted_,
+            dense_positive_mask=positive_mask,
+        )
+
+    # Filter
+    if filtering_necessary:
+        batch_filtered_scores_of_corrupted = filter_scores_(
+            scores=batch_scores_of_corrupted,
+            filter_batch=positive_filter,
+        )
+
+        # The scores for the true triples have to be rewritten to the scores tensor
+        batch_filtered_scores_of_corrupted[
+            torch.arange(0, batch.shape[0]),
+            batch[:, column],
+        ] = batch_scores_of_true
+
+        # Restrict to entities of interest
+        if restrict_entities_to is not None:
+            batch_filtered_scores_of_corrupted = batch_filtered_scores_of_corrupted[:, restrict_entities_to]
+
+        # Evaluate metrics on these *filtered* scores
+        for filtered_evaluator in filtered_evaluators:
+            if column == 2:  # tail scores
+                process = filtered_evaluator.process_tail_scores_
+            else:
+                process = filtered_evaluator.process_head_scores_
+            process(
+                hrt_batch=batch,
+                true_scores=batch_scores_of_true[:, None],
+                scores=batch_filtered_scores_of_corrupted,
+            )
+
+    return relation_filter
