@@ -8,7 +8,7 @@ import json
 import logging
 import os
 from dataclasses import dataclass
-from typing import Any, Dict, Mapping, Optional, Type, Union
+from typing import Any, Collection, Dict, Mapping, Optional, Type, Union
 
 import torch
 from optuna import Study, Trial, create_study
@@ -18,16 +18,18 @@ from optuna.storages import BaseStorage
 
 from .pruners import get_pruner_cls
 from .samplers import get_sampler_cls
-from ..datasets.base import DataSet
+from ..constants import USER_DEFINED_CODE
+from ..datasets import get_dataset, has_dataset
+from ..datasets.base import Dataset
 from ..evaluation import Evaluator, get_evaluator_cls
-from ..losses import Loss, _LOSS_SUFFIX, get_loss_cls, losses_hpo_defaults
-from ..models import get_model_cls
-from ..models.base import Model
+from ..losses import Loss, _LOSS_SUFFIX, get_loss_cls
+from ..models import Model, get_model_cls
 from ..optimizers import Optimizer, get_optimizer_cls, optimizers_hpo_defaults
 from ..pipeline import pipeline, replicate_pipeline_from_config
 from ..regularizers import Regularizer, get_regularizer_cls
 from ..sampling import NegativeSampler, get_negative_sampler_cls
 from ..stoppers import EarlyStopper, Stopper, get_stopper_cls
+from ..trackers import ResultTracker, get_result_tracker_cls
 from ..training import SLCWATrainingLoop, TrainingLoop, get_training_loop_cls
 from ..triples import TriplesFactory
 from ..utils import (
@@ -52,19 +54,22 @@ STOPPED_EPOCH_KEY = 'stopped_epoch'
 class Objective:
     """A dataclass containing all of the information to make an objective function."""
 
-    dataset: Union[None, str, Type[DataSet]]  # 1.
+    dataset: Union[None, str, Type[Dataset]]  # 1.
     model: Type[Model]  # 2.
     loss: Type[Loss]  # 3.
     regularizer: Type[Regularizer]  # 4.
     optimizer: Type[Optimizer]  # 5.
     training_loop: Type[TrainingLoop]  # 6.
     evaluator: Type[Evaluator]  # 8.
+    result_tracker: Type[ResultTracker]  # 9.
 
     # 1. Dataset
     dataset_kwargs: Optional[Mapping[str, Any]] = None
-    training_triples_factory: Optional[TriplesFactory] = None
-    testing_triples_factory: Optional[TriplesFactory] = None
-    validation_triples_factory: Optional[TriplesFactory] = None
+    training: Union[None, TriplesFactory, str] = None
+    testing: Union[None, TriplesFactory, str] = None
+    validation: Union[None, TriplesFactory, str] = None
+    evaluation_entity_whitelist: Optional[Collection[str]] = None
+    evaluation_relation_whitelist: Optional[Collection[str]] = None
     # 2. Model
     model_kwargs: Optional[Mapping[str, Any]] = None
     model_kwargs_ranges: Optional[Mapping[str, Any]] = None
@@ -89,10 +94,8 @@ class Objective:
     # 8. Evaluation
     evaluator_kwargs: Optional[Mapping[str, Any]] = None
     evaluation_kwargs: Optional[Mapping[str, Any]] = None
-    # 9. MLFlow
-    mlflow_tracking_uri: Optional[str] = None
-    mlflow_experiment_id: Optional[int] = None
-    mlflow_experiment_name: Optional[str] = None
+    # 9. Trackers
+    result_tracker_kwargs: Optional[Mapping[str, Any]] = None
     # Misc.
     metric: str = None
     device: Union[None, str, torch.device] = None
@@ -102,16 +105,13 @@ class Objective:
     def _update_stopper_callbacks(stopper_kwargs: Dict[str, Any], trial: Trial) -> None:
         """Make a subclass of the EarlyStopper that reports to the trial."""
 
-        def _continue_callback(early_stopper: EarlyStopper, result: Union[float, int]) -> None:
-            last_epoch = early_stopper.number_results * early_stopper.frequency
-            trial.report(result, step=last_epoch)
+        def _result_callback(_early_stopper: EarlyStopper, result: Union[float, int], epoch: int) -> None:
+            trial.report(result, step=epoch)
 
-        def _stopped_callback(early_stopper: EarlyStopper, result: Union[float, int]) -> None:
-            current_epoch = (1 + early_stopper.number_results) * early_stopper.frequency
-            trial.set_user_attr(STOPPED_EPOCH_KEY, int(current_epoch))
-            trial.report(result)  # don't include a step because it's over
+        def _stopped_callback(_early_stopper: EarlyStopper, _result: Union[float, int], epoch: int) -> None:
+            trial.set_user_attr(STOPPED_EPOCH_KEY, epoch)
 
-        for key, callback in zip(('continue_callbacks', 'stopped_callbacks'), (_continue_callback, _stopped_callback)):
+        for key, callback in zip(('result_callbacks', 'stopped_callbacks'), (_result_callback, _stopped_callback)):
             stopper_kwargs.setdefault(key, []).append(callback)
 
     def __call__(self, trial: Trial) -> Optional[float]:
@@ -133,11 +133,18 @@ class Objective:
             kwargs=self.model_kwargs,
             kwargs_ranges=self.model_kwargs_ranges,
         )
+
+        try:
+            loss_default_kwargs_ranges = self.loss.hpo_default
+        except AttributeError:
+            logger.warning('using a loss function with no hpo_default field: %s', self.loss)
+            loss_default_kwargs_ranges = {}
+
         # 3. Loss
         _loss_kwargs = _get_kwargs(
             trial=trial,
             prefix='loss',
-            default_kwargs_ranges=losses_hpo_defaults[self.loss],
+            default_kwargs_ranges=loss_default_kwargs_ranges,
             kwargs=self.loss_kwargs,
             kwargs_ranges=self.loss_kwargs_ranges,
         )
@@ -186,9 +193,11 @@ class Objective:
                 # 1. Dataset
                 dataset=self.dataset,
                 dataset_kwargs=self.dataset_kwargs,
-                training_triples_factory=self.training_triples_factory,
-                testing_triples_factory=self.testing_triples_factory,
-                validation_triples_factory=self.validation_triples_factory,
+                training=self.training,
+                testing=self.testing,
+                validation=self.validation,
+                evaluation_entity_whitelist=self.evaluation_entity_whitelist,
+                evaluation_relation_whitelist=self.evaluation_relation_whitelist,
                 # 2. Model
                 model=self.model,
                 model_kwargs=_model_kwargs,
@@ -214,10 +223,9 @@ class Objective:
                 evaluator=self.evaluator,
                 evaluator_kwargs=self.evaluator_kwargs,
                 evaluation_kwargs=self.evaluation_kwargs,
-                # 9. MLFlow
-                mlflow_tracking_uri=self.mlflow_tracking_uri,
-                mlflow_experiment_id=self.mlflow_experiment_id,
-                mlflow_experiment_name=self.mlflow_experiment_name,
+                # 9. Tracker
+                result_tracker=self.result_tracker,
+                result_tracker_kwargs=self.result_tracker_kwargs,
                 # Misc.
                 use_testing_data=False,  # use validation set during HPO!
                 device=self.device,
@@ -266,12 +274,14 @@ class HpoPipelineResult(Result):
                 pipeline_config[k] = v
 
         for field in dataclasses.fields(self.objective):
-            if not field.name.endswith('_kwargs') or field.name in {'metric'}:
+            field_value = getattr(self.objective, field.name)
+            if not field_value:
                 continue
-            field_kwargs = getattr(self.objective, field.name)
-            if field_kwargs:
-                logger.debug(f'saving pre-specified field in pipeline config: {field.name}={field_kwargs}')
-                pipeline_config[field.name] = field_kwargs
+            if field.name.endswith('_kwargs'):
+                logger.debug(f'saving pre-specified field in pipeline config: {field.name}={field_value}')
+                pipeline_config[field.name] = field_value
+            elif field.name in {'training', 'testing', 'validation'}:
+                pipeline_config[field.name] = field_value if isinstance(field_value, str) else USER_DEFINED_CODE
 
         for k, v in self.study.best_params.items():
             sk, ssk = k.split('.')
@@ -294,7 +304,6 @@ class HpoPipelineResult(Result):
                 f' early stopping will now switch it to {int(stopped_epoch)}'
             )
             pipeline_config['training_kwargs']['num_epochs'] = int(stopped_epoch)
-
         return dict(metadata=metadata, pipeline=pipeline_config)
 
     def save_to_directory(self, directory: str, **kwargs) -> None:
@@ -404,11 +413,13 @@ def hpo_pipeline_from_config(config: Mapping[str, Any], **kwargs) -> HpoPipeline
 def hpo_pipeline(
     *,
     # 1. Dataset
-    dataset: Union[None, str, Type[DataSet]],
+    dataset: Union[None, str, Dataset, Type[Dataset]] = None,
     dataset_kwargs: Optional[Mapping[str, Any]] = None,
-    training_triples_factory: Optional[TriplesFactory] = None,
-    testing_triples_factory: Optional[TriplesFactory] = None,
-    validation_triples_factory: Optional[TriplesFactory] = None,
+    training: Union[None, str, TriplesFactory] = None,
+    testing: Union[None, str, TriplesFactory] = None,
+    validation: Union[None, str, TriplesFactory] = None,
+    evaluation_entity_whitelist: Optional[Collection[str]] = None,
+    evaluation_relation_whitelist: Optional[Collection[str]] = None,
     # 2. Model
     model: Union[str, Type[Model]],
     model_kwargs: Optional[Mapping[str, Any]] = None,
@@ -440,10 +451,9 @@ def hpo_pipeline(
     evaluator_kwargs: Optional[Mapping[str, Any]] = None,
     evaluation_kwargs: Optional[Mapping[str, Any]] = None,
     metric: Optional[str] = None,
-    # MLFlow
-    mlflow_tracking_uri: Optional[str] = None,
-    mlflow_experiment_id: Optional[int] = None,
-    mlflow_experiment_name: Optional[str] = None,
+    # 9. Tracking
+    result_tracker: Union[None, str, Type[ResultTracker]] = None,
+    result_tracker_kwargs: Optional[Mapping[str, Any]] = None,
     # 6. Misc
     device: Union[None, str, torch.device] = None,
     #  Optuna Study Settings
@@ -464,16 +474,25 @@ def hpo_pipeline(
     """Train a model on the given dataset.
 
     :param dataset:
-        The name of the dataset (a key from :data:`pykeen.datasets.datasets`) or the :class:`pykeen.datasets.DataSet`
-        instance. Alternatively, the ``training_triples_factory`` and ``testing_triples_factory`` can be specified.
+        The name of the dataset (a key from :data:`pykeen.datasets.datasets`) or the :class:`pykeen.datasets.Dataset`
+        instance. Alternatively, the training triples factory (``training``), testing triples factory (``testing``),
+        and validation triples factory (``validation``; optional) can be specified.
     :param dataset_kwargs:
         The keyword arguments passed to the dataset upon instantiation
-    :param training_triples_factory:
-        A triples factory with training instances if a a dataset was not specified
-    :param testing_triples_factory:
-        A triples factory with training instances if a dataset was not specified
-    :param validation_triples_factory:
-        A triples factory with validation instances if a dataset was not specified
+    :param training:
+        A triples factory with training instances or path to the training file if a a dataset was not specified
+    :param testing:
+        A triples factory with test instances or path to the test file if a dataset was not specified
+    :param validation:
+        A triples factory with validation instances or path to the validation file if a dataset was not specified
+    :param evaluation_entity_whitelist:
+        Optional restriction of evaluation to triples containing *only* these entities. Useful if the downstream task
+        is only interested in certain entities, but the relational patterns with other entities improve the entity
+        embedding quality. Passed to :func:`pykeen.pipeline.pipeline`.
+    :param evaluation_relation_whitelist:
+        Optional restriction of evaluation to triples containing *only* these relations. Useful if the downstream task
+        is only interested in certain relation, but the relational patterns with other relations improve the entity
+        embedding quality. Passed to :func:`pykeen.pipeline.pipeline`.
 
     :param model:
         The name of the model or the model class to pass to :func:`pykeen.pipeline.pipeline`
@@ -537,14 +556,10 @@ def hpo_pipeline(
     :param evaluation_kwargs:
         Keyword arguments to pass to the evaluator's evaluate function on call
 
-    :param mlflow_tracking_uri:
-        The MLFlow tracking URL. If None is given, MLFlow is not used to track results.
-    :param mlflow_experiment_id:
-        The experiment ID. If given, this has to be the ID of an existing experiment in MFLow. Has priority over
-        experiment_name. Only effective if mlflow_tracking_uri is not None.
-    :param mlflow_experiment_name:
-        The experiment name. If this experiment name exists, add the current run to this experiment. Otherwise
-        create an experiment of the given name. Only effective if mlflow_tracking_uri is not None.
+    :param result_tracker:
+        The ResultsTracker class or name
+    :param result_tracker_kwargs:
+        The keyword arguments passed to the results tracker on instantiation
 
     :param metric:
         The metric to optimize over. Defaults to ``adjusted_mean_rank``.
@@ -579,10 +594,15 @@ def hpo_pipeline(
     study.set_user_attr('pykeen_version', get_version())
     study.set_user_attr('pykeen_git_hash', get_git_hash())
     # 1. Dataset
-    # FIXME difference between dataset class and string
-    # FIXME how to handle if dataset or factories were set? Should have been
-    #  part of https://github.com/mali-git/POEM_develop/pull/483
-    study.set_user_attr('dataset', dataset)
+    _set_study_dataset(
+        study=study,
+        dataset=dataset,
+        dataset_kwargs=dataset_kwargs,
+        training=training,
+        testing=testing,
+        validation=validation,
+    )
+
     # 2. Model
     model: Type[Model] = get_model_cls(model)
     study.set_user_attr('model', normalize_string(model.__name__))
@@ -628,13 +648,18 @@ def hpo_pipeline(
     study.set_user_attr('metric', metric)
     logger.info(f'Attempting to {direction} {metric}')
 
+    # 9. Tracking
+    result_tracker: Type[ResultTracker] = get_result_tracker_cls(result_tracker)
+
     objective = Objective(
         # 1. Dataset
         dataset=dataset,
         dataset_kwargs=dataset_kwargs,
-        training_triples_factory=training_triples_factory,
-        testing_triples_factory=testing_triples_factory,
-        validation_triples_factory=validation_triples_factory,
+        training=training,
+        testing=testing,
+        validation=validation,
+        evaluation_entity_whitelist=evaluation_entity_whitelist,
+        evaluation_relation_whitelist=evaluation_relation_whitelist,
         # 2. Model
         model=model,
         model_kwargs=model_kwargs,
@@ -665,10 +690,9 @@ def hpo_pipeline(
         evaluator=evaluator,
         evaluator_kwargs=evaluator_kwargs,
         evaluation_kwargs=evaluation_kwargs,
-        # 9. MLFlow
-        mlflow_tracking_uri=mlflow_tracking_uri,
-        mlflow_experiment_id=mlflow_experiment_id,
-        mlflow_experiment_name=mlflow_experiment_name,
+        # 9. Tracker
+        result_tracker=result_tracker,
+        result_tracker_kwargs=result_tracker_kwargs,
         # Optuna Misc.
         metric=metric,
         save_model_directory=save_model_directory,
@@ -775,3 +799,26 @@ def suggest_discrete_power_two_int(trial: Trial, name, low, high) -> int:
         raise Exception(f"Upper bound {high} is not greater than lower bound {low}.")
     choices = [2 ** i for i in range(low, high + 1)]
     return trial.suggest_categorical(name=name, choices=choices)
+
+
+def _set_study_dataset(
+    study: Study,
+    *,
+    dataset: Union[None, str, Dataset, Type[Dataset]] = None,
+    dataset_kwargs: Optional[Mapping[str, Any]] = None,
+    training: Union[None, str, TriplesFactory] = None,
+    testing: Union[None, str, TriplesFactory] = None,
+    validation: Union[None, str, TriplesFactory] = None,
+):
+    if (
+        (isinstance(dataset, str) and has_dataset(dataset))
+        or isinstance(dataset, Dataset)
+        or (isinstance(dataset, type) and issubclass(dataset, Dataset))
+    ):
+        dataset_name = get_dataset(dataset=dataset).get_normalized_name()
+        study.set_user_attr('dataset', dataset_name)
+    else:
+        study.set_user_attr('dataset', USER_DEFINED_CODE)
+        study.set_user_attr('training', training if isinstance(training, str) else USER_DEFINED_CODE)
+        study.set_user_attr('testing', testing if isinstance(testing, str) else USER_DEFINED_CODE)
+        study.set_user_attr('validation', validation if isinstance(validation, str) else USER_DEFINED_CODE)
