@@ -14,11 +14,11 @@ about each and see the reference on how to use them specifically. Don't worry, i
 the tutorial, the :func:`pykeen.pipeline.pipeline` function will take care of everything for you.
 
 >>> from pykeen.pipeline import pipeline
->>> result = pipeline(
+>>> pipeline_result = pipeline(
 ...     dataset='Nations',
 ...     model='TransE',
 ... )
->>> result.save_to_directory('nations_transe')
+>>> pipeline_result.save_to_directory('nations_transe')
 
 The results are returned in a :class:`pykeen.pipeline.PipelineResult` instance, which has
 attributes for the trained model, the training loop, and the evaluation.
@@ -29,60 +29,60 @@ could be used as in:
 
 >>> from pykeen.pipeline import pipeline
 >>> from pykeen.models import TransE
->>> result = pipeline(
+>>> pipeline_result = pipeline(
 ...     dataset='Nations',
 ...     model=TransE,
 ... )
->>> result.save_to_directory('nations_transe')
+>>> pipeline_result.save_to_directory('nations_transe')
 
-In this example, the data set was given as a string. A list of available data sets can be found in
-:mod:`pykeen.datasets`. Alternatively, the instance of the :class:`pykeen.datasets.DataSet` could be
+In this example, the dataset was given as a string. A list of available datasets can be found in
+:mod:`pykeen.datasets`. Alternatively, the instance of the :class:`pykeen.datasets.Dataset` could be
 used as in:
 
 >>> from pykeen.pipeline import pipeline
 >>> from pykeen.models import TransE
 >>> from pykeen.datasets import Nations
->>> result = pipeline(
+>>> pipeline_result = pipeline(
 ...     dataset=Nations,
 ...     model=TransE,
 ... )
->>> result.save_to_directory('nations_transe')
+>>> pipeline_result.save_to_directory('nations_transe')
 
 In each of the previous three examples, the training approach, optimizer, and evaluation scheme
 were omitted. By default, the stochastic local closed world assumption (sLCWA) training approach is used in training.
 This can be explicitly given as a string:
 
 >>> from pykeen.pipeline import pipeline
->>> result = pipeline(
+>>> pipeline_result = pipeline(
 ...     dataset='Nations',
 ...     model='TransE',
 ...     training_loop='sLCWA',
 ... )
->>> result.save_to_directory('nations_transe')
+>>> pipeline_result.save_to_directory('nations_transe')
 
 Alternatively, the local closed world assumption (LCWA) training approach can be given with ``'LCWA'``.
 No additional configuration is necessary, but it's worth reading up on the differences between these training
 approaches.
 
 >>> from pykeen.pipeline import pipeline
->>> result = pipeline(
+>>> pipeline_result = pipeline(
 ...     dataset='Nations',
 ...     model='TransE',
 ...     training_loop='LCWA',
 ... )
->>> result.save_to_directory('nations_transe')
+>>> pipeline_result.save_to_directory('nations_transe')
 
 One of these differences is that the sLCWA relies on *negative sampling*. The type of negative sampling
 can be given as in:
 
 >>> from pykeen.pipeline import pipeline
->>> result = pipeline(
+>>> pipeline_result = pipeline(
 ...     dataset='Nations',
 ...     model='TransE',
 ...     training_loop='sLCWA',
 ...     negative_sampler='basic',
 ... )
->>> result.save_to_directory('nations_transe')
+>>> pipeline_result.save_to_directory('nations_transe')
 
 In this example, the negative sampler was given as a string. A list of available negative samplers
 can be found in :mod:`pykeen.sampling`. Alternatively, the class corresponding to the implementation
@@ -90,13 +90,13 @@ of the negative sampler could be used as in:
 
 >>> from pykeen.pipeline import pipeline
 >>> from pykeen.sampling import BasicNegativeSampler
->>> result = pipeline(
+>>> pipeline_result = pipeline(
 ...     dataset='Nations',
 ...     model='TransE',
 ...     training_loop='sLCWA',
 ...     negative_sampler=BasicNegativeSampler,
 ... )
->>> result.save_to_directory('nations_transe')
+>>> pipeline_result.save_to_directory('nations_transe')
 
 .. warning ::
 
@@ -107,12 +107,12 @@ The type of evaluation perfomed can be specified with the ``evaluator`` keyword.
 rank-based evaluation is used. It can be given explictly as in:
 
 >>> from pykeen.pipeline import pipeline
->>> result = pipeline(
+>>> pipeline_result = pipeline(
 ...     dataset='Nations',
 ...     model='TransE',
 ...     evaluator='RankBasedEvaluator',
 ... )
->>> result.save_to_directory('nations_transe')
+>>> pipeline_result.save_to_directory('nations_transe')
 
 In this example, the evaluator string. A list of available evaluators can be found in
 :mod:`pykeen.evaluation`. Alternatively, the class corresponding to the implementation
@@ -120,23 +120,23 @@ of the evaluator could be used as in:
 
 >>> from pykeen.pipeline import pipeline
 >>> from pykeen.evaluation import RankBasedEvaluator
->>> result = pipeline(
+>>> pipeline_result = pipeline(
 ...     dataset='Nations',
 ...     model='TransE',
 ...     evaluator=RankBasedEvaluator,
 ... )
->>> result.save_to_directory('nations_transe')
+>>> pipeline_result.save_to_directory('nations_transe')
 
 PyKEEN implements early stopping, which can be turned on with the ``stopper`` keyword
 argument as in:
 
 >>> from pykeen.pipeline import pipeline
->>> result = pipeline(
+>>> pipeline_result = pipeline(
 ...     dataset='Nations',
 ...     model='TransE',
 ...     stopper='early',
 ... )
->>> result.save_to_directory('nations_transe')
+>>> pipeline_result.save_to_directory('nations_transe')
 
 Deeper Configuration
 ~~~~~~~~~~~~~~~~~~~~
@@ -150,11 +150,13 @@ Arguments for the model can be given as a dictionary using ``model_kwargs``.
 ...         scoring_fct_norm=2,
 ...     ),
 ... )
->>> result.save_to_directory('nations_transe')
+>>> pipeline_result.save_to_directory('nations_transe')
 
 The entries in ``model_kwargs`` correspond to the arguments given to :func:`pykeen.models.TransE.__init__`. For a
 complete listing of models, see :mod:`pykeen.models`, where there are links to the reference for each
-model that explain what kwargs are possible.
+model that explain what kwargs are possible. Each model's default hyper-parameters were chosen based on the
+best reported values from the paper originally publishing the model unless otherwise noted on the model's
+reference page.
 
 Because the pipeline takes care of looking up classes and instantiating them,
 there are several other parameters to :func:`pykeen.pipeline.pipeline` that
@@ -168,31 +170,34 @@ import ftplib
 import json
 import logging
 import os
-import random
+import pathlib
+import pickle
 import time
 from dataclasses import dataclass, field
-from typing import Any, Dict, Iterable, List, Mapping, Optional, Type, Union
+from pathlib import Path
+from typing import Any, Collection, Dict, Iterable, List, Mapping, Optional, Set, Type, Union
 
 import pandas as pd
 import torch
 from torch.optim.optimizer import Optimizer
 
+from .constants import PYKEEN_CHECKPOINTS, USER_DEFINED_CODE
 from .datasets import get_dataset
-from .datasets.base import DataSet
+from .datasets.base import Dataset
 from .evaluation import Evaluator, MetricResults, get_evaluator_cls
-from .losses import Loss, get_loss_cls
-from .models import get_model_cls
-from .models.base import Model
+from .losses import Loss, _LOSS_SUFFIX, get_loss_cls
+from .models import Model, get_model_cls
+from .nn import Embedding
 from .optimizers import get_optimizer_cls
 from .regularizers import Regularizer, get_regularizer_cls
 from .sampling import NegativeSampler, get_negative_sampler_cls
 from .stoppers import EarlyStopper, Stopper, get_stopper_cls
-from .trackers import MLFlowResultTracker, ResultTracker
+from .trackers import ResultTracker, get_result_tracker_cls
 from .training import SLCWATrainingLoop, TrainingLoop, get_training_loop_cls
 from .triples import TriplesFactory
 from .utils import (
-    NoRandomSeedNecessary, Result, ensure_ftp_directory, fix_dataclass_init_docs, get_json_bytes_io, get_model_io,
-    resolve_device, set_random_seed,
+    Result, ensure_ftp_directory, fix_dataclass_init_docs, get_json_bytes_io, get_model_io, normalize_string,
+    random_non_negative_int, resolve_device, set_random_seed,
 )
 from .version import get_git_hash, get_version
 
@@ -207,6 +212,8 @@ __all__ = [
 
 logger = logging.getLogger(__name__)
 
+REDUCER_RELATION_WHITELIST = {'PCA'}
+
 
 @fix_dataclass_init_docs
 @dataclass
@@ -251,16 +258,148 @@ class PipelineResult(Result):
         """The title of the experiment."""
         return self.metadata.get('title')
 
-    def plot_losses(self):
+    def plot_losses(self, ax=None):
         """Plot the losses per epoch."""
-        import matplotlib.pyplot as plt
+        if ax is None:
+            import matplotlib.pyplot as plt
+            ax = plt.gca()
+
         import seaborn as sns
-        sns.set()
-        plt.ylabel('Loss')
-        plt.xlabel('Epoch')
-        if self.title is not None:
-            plt.title(self.title)
-        return sns.lineplot(x=range(len(self.losses)), y=self.losses)
+        sns.set_style('darkgrid')
+
+        rv = sns.lineplot(x=range(len(self.losses)), y=self.losses, ax=ax)
+
+        loss_name = normalize_string(self.model.loss.__class__.__name__, suffix=_LOSS_SUFFIX)
+        ax.set_ylabel(f'{loss_name} Loss')
+        ax.set_xlabel('Epoch')
+        ax.set_title(self.title if self.title is not None else 'Losses Plot')
+        return rv
+
+    def plot_er(  # noqa: C901
+        self,
+        model: Optional[str] = None,
+        margin: float = 0.4,
+        ax=None,
+        entities: Optional[Set[str]] = None,
+        relations: Optional[Set[str]] = None,
+        apply_limits: bool = True,
+        plot_entities: bool = True,
+        plot_relations: Optional[bool] = None,
+        annotation_x_offset: float = 0.02,
+        annotation_y_offset: float = 0.03,
+        **kwargs,
+    ):
+        """Plot the reduced entities and relation vectors in 2D.
+
+        :param model: The dimensionality reduction model from :mod:`sklearn`. Defaults to PCA.
+            Can also use KPCA, GRP, SRP, TSNE, LLE, ISOMAP, MDS, or SE.
+        :param kwargs: The keyword arguments passed to `__init__()` of
+            the reducer class (e.g., PCA, TSNE)
+        :param plot_relations: By default, this is only enabled on translational distance models
+            like :class:`pykeen.models.TransE`.
+
+        .. warning::
+
+            Plotting relations and entities on the same plot is only
+            meaningful for translational distance models like TransE.
+        """
+        if not plot_entities and not plot_relations:
+            raise ValueError
+
+        if plot_relations is None:  # automatically set to true for translational models, false otherwise
+            plot_relations = self.model.__class__.__name__.lower().startswith('trans')
+
+        if model is None:
+            model = 'PCA'
+        reducer_cls, reducer_kwargs = _get_reducer_cls(model, **kwargs)
+        if plot_relations and reducer_cls.__name__ not in REDUCER_RELATION_WHITELIST:
+            raise ValueError(f'Can not use reducer {reducer_cls} when projecting relations. Will result in nonsense')
+        reducer = reducer_cls(n_components=2, **reducer_kwargs)
+
+        if ax is None:
+            import matplotlib.pyplot as plt
+            ax = plt.gca()
+
+        import seaborn as sns
+        sns.set_style('whitegrid')
+
+        if plot_relations and plot_entities:
+            e_embeddings, e_reduced = _reduce_embeddings(self.model.entity_embeddings, reducer, fit=True)
+            r_embeddings, r_reduced = _reduce_embeddings(self.model.relation_embeddings, reducer, fit=False)
+
+            xmax = max(r_embeddings[:, 0].max(), e_embeddings[:, 0].max()) + margin
+            xmin = min(r_embeddings[:, 0].min(), e_embeddings[:, 0].min()) - margin
+            ymax = max(r_embeddings[:, 1].max(), e_embeddings[:, 1].max()) + margin
+            ymin = min(r_embeddings[:, 1].min(), e_embeddings[:, 1].min()) - margin
+        elif plot_relations:
+            e_embeddings, e_reduced = None, False
+            r_embeddings, r_reduced = _reduce_embeddings(self.model.relation_embeddings, reducer, fit=True)
+
+            xmax = r_embeddings[:, 0].max() + margin
+            xmin = r_embeddings[:, 0].min() - margin
+            ymax = r_embeddings[:, 1].max() + margin
+            ymin = r_embeddings[:, 1].min() - margin
+        elif plot_entities:
+            e_embeddings, e_reduced = _reduce_embeddings(self.model.entity_embeddings, reducer, fit=True)
+            r_embeddings, r_reduced = None, False
+
+            xmax = e_embeddings[:, 0].max() + margin
+            xmin = e_embeddings[:, 0].min() - margin
+            ymax = e_embeddings[:, 1].max() + margin
+            ymin = e_embeddings[:, 1].min() - margin
+        else:
+            raise ValueError  # not even possible
+
+        if not e_reduced and not r_reduced:
+            subtitle = ''
+        elif reducer_kwargs:
+            subtitle = ", ".join("=".join(item) for item in reducer_kwargs.items())
+            subtitle = f' using {reducer_cls.__name__} ({subtitle})'
+        else:
+            subtitle = f' using {reducer_cls.__name__}'
+
+        if plot_entities:
+            entity_id_to_label = self.model.triples_factory.entity_id_to_label
+            for entity_id, entity_reduced_embedding in enumerate(e_embeddings):
+                entity_label = entity_id_to_label[entity_id]
+                if entities and entity_label not in entities:
+                    continue
+                x, y = entity_reduced_embedding
+                ax.scatter(x, y, color='black')
+                ax.annotate(entity_label, (x + annotation_x_offset, y + annotation_y_offset))
+
+        if plot_relations:
+            relation_id_to_label = self.model.triples_factory.relation_id_to_label
+            for relation_id, relation_reduced_embedding in enumerate(r_embeddings):
+                relation_label = relation_id_to_label[relation_id]
+                if relations and relation_label not in relations:
+                    continue
+                x, y = relation_reduced_embedding
+                ax.arrow(0, 0, x, y, color='black')
+                ax.annotate(relation_label, (x + annotation_x_offset, y + annotation_y_offset))
+
+        if plot_entities and plot_relations:
+            ax.set_title(f'Entity/Relation Plot{subtitle}')
+        elif plot_entities:
+            ax.set_title(f'Entity Plot{subtitle}')
+        elif plot_relations:
+            ax.set_title(f'Relation Plot{subtitle}')
+
+        if apply_limits:
+            ax.set_xlim([xmin, xmax])
+            ax.set_ylim([ymin, ymax])
+
+        return ax
+
+    def plot(self, er_kwargs: Optional[Mapping[str, str]] = None, figsize=(10, 4)):
+        """Plot all plots."""
+        import matplotlib.pyplot as plt
+        fig, (lax, rax) = plt.subplots(1, 2, figsize=figsize)
+
+        self.plot_losses(ax=lax)
+        self.plot_er(ax=rax, **(er_kwargs or {}))
+
+        plt.tight_layout()
 
     def save_model(self, path: str) -> None:
         """Save the trained model to the given path using :func:`torch.save`.
@@ -270,7 +409,7 @@ class PipelineResult(Result):
 
         The model contains within it the triples factory that was used for training.
         """
-        torch.save(self.model, path)
+        torch.save(self.model, path, pickle_protocol=pickle.HIGHEST_PROTOCOL)
 
     def _get_results(self) -> Mapping[str, Any]:
         results = dict(
@@ -285,7 +424,12 @@ class PipelineResult(Result):
             results['stopper'] = self.stopper.get_summary_dict()
         return results
 
-    def save_to_directory(self, directory: str, save_metadata: bool = True, save_replicates: bool = True) -> None:
+    def save_to_directory(
+        self,
+        directory: Union[str, Path],
+        save_metadata: bool = True,
+        save_replicates: bool = True,
+    ) -> None:
         """Save all artifacts in the given directory."""
         os.makedirs(directory, exist_ok=True)
 
@@ -387,6 +531,49 @@ class PipelineResult(Result):
         s3.upload_fileobj(get_model_io(self.model), bucket, model_path)
 
 
+def _reduce_embeddings(embedding: Embedding, reducer, fit: bool = False):
+    embeddings_numpy = embedding(indices=None).detach().cpu().numpy()
+    if embeddings_numpy.shape[1] == 2:
+        logger.debug('not reducing entity embeddings, already dim=2')
+        return embeddings_numpy, False
+    elif fit:
+        return reducer.fit_transform(embeddings_numpy), True
+    else:
+        return reducer.transform(embeddings_numpy), True
+
+
+def _get_reducer_cls(model: str, **kwargs):
+    """Get the model class by name and default kwargs.
+
+    :param model: The name of the model. Can choose from: PCA, KPCA, GRP,
+        SRP, TSNE, LLE, ISOMAP, MDS, or SE.
+    :param kwargs:
+    :return:
+    """
+    if model.upper() == 'PCA':
+        from sklearn.decomposition import PCA as Reducer  # noqa:N811
+    elif model.upper() == 'KPCA':
+        kwargs.setdefault('kernel', 'rbf')
+        from sklearn.decomposition import KernelPCA as Reducer
+    elif model.upper() == 'GRP':
+        from sklearn.random_projection import GaussianRandomProjection as Reducer
+    elif model.upper() == 'SRP':
+        from sklearn.random_projection import SparseRandomProjection as Reducer
+    elif model.upper() in {'T-SNE', 'TSNE'}:
+        from sklearn.manifold import TSNE as Reducer  # noqa:N811
+    elif model.upper() in {'LLE', 'LOCALLYLINEAREMBEDDING'}:
+        from sklearn.manifold import LocallyLinearEmbedding as Reducer
+    elif model.upper() == 'ISOMAP':
+        from sklearn.manifold import Isomap as Reducer
+    elif model.upper() in {'MDS', 'MULTIDIMENSIONALSCALING'}:
+        from sklearn.manifold import MDS as Reducer  # noqa:N811
+    elif model.upper() in {'SE', 'SPECTRAL', 'SPECTRALEMBEDDING'}:
+        from sklearn.manifold import SpectralEmbedding as Reducer
+    else:
+        raise ValueError(f'invalid dimensionality reduction model: {model}')
+    return Reducer, kwargs
+
+
 def replicate_pipeline_from_path(
     path: str,
     directory: str,
@@ -447,7 +634,8 @@ def replicate_pipeline_from_config(
 
 def _iterate_moved(pipeline_results: Iterable[PipelineResult]):
     for pipeline_result in pipeline_results:
-        pipeline_result.model.to_cpu_()
+        pipeline_result.model.device = resolve_device('cpu')
+        pipeline_result.model.to_device_()
         yield pipeline_result
 
 
@@ -489,32 +677,27 @@ def save_pipeline_results_to_directory(
 
 def pipeline_from_path(
     path: str,
-    mlflow_tracking_uri: Optional[str] = None,
     **kwargs,
 ) -> PipelineResult:
     """Run the pipeline with configuration in a JSON file at the given path.
 
     :param path: The path to an experiment JSON file
-    :param mlflow_tracking_uri: The URL of the MLFlow tracking server. If None, do not use MLFlow for result tracking.
     """
     with open(path) as file:
         config = json.load(file)
     return pipeline_from_config(
         config=config,
-        mlflow_tracking_uri=mlflow_tracking_uri,
         **kwargs,
     )
 
 
 def pipeline_from_config(
     config: Mapping[str, Any],
-    mlflow_tracking_uri: Optional[str] = None,
     **kwargs,
 ) -> PipelineResult:
     """Run the pipeline with a configuration dictionary.
 
     :param config: The experiment configuration dictionary
-    :param mlflow_tracking_uri: The URL of the MLFlow tracking server. If None, do not use MLFlow for result tracking.
     """
     metadata, pipeline_kwargs = config['metadata'], config['pipeline']
     title = metadata.get('title')
@@ -522,7 +705,6 @@ def pipeline_from_config(
         logger.info(f'Running: {title}')
 
     return pipeline(
-        mlflow_tracking_uri=mlflow_tracking_uri,
         metadata=metadata,
         **pipeline_kwargs,
         **kwargs,
@@ -532,11 +714,13 @@ def pipeline_from_config(
 def pipeline(  # noqa: C901
     *,
     # 1. Dataset
-    dataset: Union[None, str, Type[DataSet]] = None,
+    dataset: Union[None, str, Dataset, Type[Dataset]] = None,
     dataset_kwargs: Optional[Mapping[str, Any]] = None,
-    training_triples_factory: Optional[TriplesFactory] = None,
-    testing_triples_factory: Optional[TriplesFactory] = None,
-    validation_triples_factory: Optional[TriplesFactory] = None,
+    training: Union[None, TriplesFactory, str] = None,
+    testing: Union[None, TriplesFactory, str] = None,
+    validation: Union[None, TriplesFactory, str] = None,
+    evaluation_entity_whitelist: Optional[Collection[str]] = None,
+    evaluation_relation_whitelist: Optional[Collection[str]] = None,
     # 2. Model
     model: Union[str, Type[Model]],
     model_kwargs: Optional[Mapping[str, Any]] = None,
@@ -562,11 +746,11 @@ def pipeline(  # noqa: C901
     evaluator: Union[None, str, Type[Evaluator]] = None,
     evaluator_kwargs: Optional[Mapping[str, Any]] = None,
     evaluation_kwargs: Optional[Mapping[str, Any]] = None,
-    # 9. MLFlow
-    mlflow_tracking_uri: Optional[str] = None,
-    mlflow_experiment_id: Optional[int] = None,
-    mlflow_experiment_name: Optional[str] = None,
+    # 9. Tracking
+    result_tracker: Union[None, str, Type[ResultTracker]] = None,
+    result_tracker_kwargs: Optional[Mapping[str, Any]] = None,
     # Misc
+    automatic_memory_optimization: bool = True,
     metadata: Optional[Dict[str, Any]] = None,
     device: Union[None, str, torch.device] = None,
     random_seed: Optional[int] = None,
@@ -575,16 +759,25 @@ def pipeline(  # noqa: C901
     """Train and evaluate a model.
 
     :param dataset:
-        The name of the dataset (a key from :data:`pykeen.datasets.datasets`) or the :class:`pykeen.datasets.DataSet`
-        instance. Alternatively, the ``training_triples_factory`` and ``testing_triples_factory`` can be specified.
+        The name of the dataset (a key from :data:`pykeen.datasets.datasets`) or the :class:`pykeen.datasets.Dataset`
+        instance. Alternatively, the training triples factory (``training``), testing triples factory (``testing``),
+        and validation triples factory (``validation``; optional) can be specified.
     :param dataset_kwargs:
         The keyword arguments passed to the dataset upon instantiation
-    :param training_triples_factory:
-        A triples factory with training instances if a a dataset was not specified
-    :param testing_triples_factory:
-        A triples factory with training instances if a dataset was not specified
-    :param validation_triples_factory:
-        A triples factory with validation instances if a dataset was not specified
+    :param training:
+        A triples factory with training instances or path to the training file if a a dataset was not specified
+    :param testing:
+        A triples factory with training instances or path to the test file if a dataset was not specified
+    :param validation:
+        A triples factory with validation instances or path to the validation file if a dataset was not specified
+    :param evaluation_entity_whitelist:
+        Optional restriction of evaluation to triples containing *only* these entities. Useful if the downstream task
+        is only interested in certain entities, but the relational patterns with other entities improve the entity
+        embedding quality.
+    :param evaluation_relation_whitelist:
+        Optional restriction of evaluation to triples containing *only* these relations. Useful if the downstream task
+        is only interested in certain relation, but the relational patterns with other relations improve the entity
+        embedding quality.
 
     :param model:
         The name of the model or the model class
@@ -634,33 +827,44 @@ def pipeline(  # noqa: C901
     :param evaluation_kwargs:
         Keyword arguments to pass to the evaluator's evaluate function on call
 
-    :param mlflow_tracking_uri:
-        The MLFlow tracking URL. If None is given, MLFlow is not used to track results.
-    :param mlflow_experiment_id:
-        The experiment ID. If given, this has to be the ID of an existing experiment in MFLow. Has priority over
-        experiment_name. Only effective if mlflow_tracking_uri is not None.
-    :param mlflow_experiment_name:
-        The experiment name. If this experiment name exists, add the current run to this experiment. Otherwise
-        create an experiment of the given name. Only effective if mlflow_tracking_uri is not None.
+    :param result_tracker:
+        The ResultsTracker class or name
+    :param result_tracker_kwargs:
+        The keyword arguments passed to the results tracker on instantiation
 
-    :param metadata: A JSON dictionary to store with the experiment
-    :param use_testing_data: If true, use the testing triples. Otherwise, use the validation triples.
-     Defaults to true - use testing triples.
+    :param metadata:
+        A JSON dictionary to store with the experiment
+    :param use_testing_data:
+        If true, use the testing triples. Otherwise, use the validation triples. Defaults to true - use testing triples.
     """
-    if random_seed is None:
-        random_seed = random.randint(0, 2 ** 32 - 1)
+    if training_kwargs is None:
+        training_kwargs = {}
+
+    # To allow resuming training from a checkpoint when using a pipeline, the pipeline needs to obtain the
+    # used random_seed to ensure reproducible results
+    checkpoint_name = training_kwargs.get('checkpoint_name')
+    if checkpoint_name is not None:
+        checkpoint_directory = pathlib.Path(training_kwargs.get('checkpoint_directory', PYKEEN_CHECKPOINTS))
+        checkpoint_directory.mkdir(parents=True, exist_ok=True)
+        checkpoint_path = checkpoint_directory / checkpoint_name
+        if checkpoint_path.is_file():
+            checkpoint_dict = torch.load(checkpoint_path)
+            random_seed = checkpoint_dict['random_seed']
+            logger.info('loaded random seed %s from checkpoint.', random_seed)
+            # We have to set clear optimizer to False since training should be continued
+            clear_optimizer = False
+        else:
+            logger.info(f"=> no training loop checkpoint file found at '{checkpoint_path}'. Creating a new file.")
+            if random_seed is None:
+                random_seed = random_non_negative_int()
+                logger.warning(f'No random seed is specified. Setting to {random_seed}.')
+    elif random_seed is None:
+        random_seed = random_non_negative_int()
         logger.warning(f'No random seed is specified. Setting to {random_seed}.')
     set_random_seed(random_seed)
 
-    # Create result store
-    if mlflow_tracking_uri is not None:
-        result_tracker = MLFlowResultTracker(
-            tracking_uri=mlflow_tracking_uri,
-            experiment_id=mlflow_experiment_id,
-            experiment_name=mlflow_experiment_name,
-        )
-    else:
-        result_tracker = ResultTracker()
+    result_tracker_cls: Type[ResultTracker] = get_result_tracker_cls(result_tracker)
+    result_tracker = result_tracker_cls(**(result_tracker_kwargs or {}))
 
     if not metadata:
         metadata = {}
@@ -671,20 +875,40 @@ def pipeline(  # noqa: C901
 
     device = resolve_device(device)
 
-    result_tracker.log_params(dict(dataset=dataset))
-
-    training_triples_factory, testing_triples_factory, validation_triples_factory = get_dataset(
+    dataset_instance: Dataset = get_dataset(
         dataset=dataset,
         dataset_kwargs=dataset_kwargs,
-        training_triples_factory=training_triples_factory,
-        testing_triples_factory=testing_triples_factory,
-        validation_triples_factory=validation_triples_factory,
+        training=training,
+        testing=testing,
+        validation=validation,
     )
+    if dataset is not None:
+        result_tracker.log_params(dict(dataset=dataset_instance.get_normalized_name()))
+    else:  # means that dataset was defined by triples factories
+        result_tracker.log_params(dict(
+            dataset=USER_DEFINED_CODE,
+            training=training if isinstance(training, str) else USER_DEFINED_CODE,
+            testing=testing if isinstance(training, str) else USER_DEFINED_CODE,
+            validation=validation if isinstance(training, str) else USER_DEFINED_CODE,
+        ))
+
+    training, testing, validation = dataset_instance.training, dataset_instance.testing, dataset_instance.validation
+    # evaluation restriction to a subset of entities/relations
+    if any(f is not None for f in (evaluation_entity_whitelist, evaluation_relation_whitelist)):
+        testing = testing.new_with_restriction(
+            entities=evaluation_entity_whitelist,
+            relations=evaluation_relation_whitelist,
+        )
+        if validation is not None:
+            validation = validation.new_with_restriction(
+                entities=evaluation_entity_whitelist,
+                relations=evaluation_relation_whitelist,
+            )
 
     if model_kwargs is None:
         model_kwargs = {}
     model_kwargs.update(preferred_device=device)
-    model_kwargs.setdefault('random_seed', NoRandomSeedNecessary)
+    model_kwargs.setdefault('random_seed', random_seed)
 
     if regularizer is not None:
         # FIXME this should never happen.
@@ -693,7 +917,6 @@ def pipeline(  # noqa: C901
             del model_kwargs['regularizer']
         regularizer_cls: Type[Regularizer] = get_regularizer_cls(regularizer)
         model_kwargs['regularizer'] = regularizer_cls(
-            device=device,
             **(regularizer_kwargs or {}),
         )
 
@@ -707,7 +930,7 @@ def pipeline(  # noqa: C901
 
     model = get_model_cls(model)
     model_instance: Model = model(
-        triples_factory=training_triples_factory,
+        triples_factory=training,
         **model_kwargs,
     )
     # Log model parameters
@@ -731,6 +954,7 @@ def pipeline(  # noqa: C901
         training_loop_instance: TrainingLoop = training_loop(
             model=model_instance,
             optimizer=optimizer_instance,
+            automatic_memory_optimization=automatic_memory_optimization,
         )
     elif training_loop is not SLCWATrainingLoop:
         raise ValueError('Can not specify negative sampler with LCWA')
@@ -738,26 +962,26 @@ def pipeline(  # noqa: C901
         negative_sampler = get_negative_sampler_cls(negative_sampler)
         result_tracker.log_params(
             params=dict(cls=negative_sampler.__name__, kwargs=negative_sampler_kwargs),
-            prefix='negative_sampler'
+            prefix='negative_sampler',
         )
         training_loop_instance: TrainingLoop = SLCWATrainingLoop(
             model=model_instance,
             optimizer=optimizer_instance,
+            automatic_memory_optimization=automatic_memory_optimization,
             negative_sampler_cls=negative_sampler,
             negative_sampler_kwargs=negative_sampler_kwargs,
         )
 
     evaluator = get_evaluator_cls(evaluator)
-    evaluator_instance: Evaluator = evaluator(
-        **(evaluator_kwargs or {}),
-    )
+
+    if evaluator_kwargs is None:
+        evaluator_kwargs = {}
+    evaluator_kwargs.setdefault('automatic_memory_optimization', automatic_memory_optimization)
+    evaluator_instance: Evaluator = evaluator(**evaluator_kwargs)
 
     if evaluation_kwargs is None:
         evaluation_kwargs = {}
 
-    if training_kwargs is None:
-        training_kwargs = {}
-
     # Stopping
     if 'stopper' in training_kwargs and stopper is not None:
         raise ValueError('Specified stopper in training_kwargs and as stopper')
@@ -776,7 +1000,7 @@ def pipeline(  # noqa: C901
     stopper: Stopper = stopper_cls(
         model=model_instance,
         evaluator=evaluator_instance,
-        evaluation_triples_factory=validation_triples_factory,
+        evaluation_triples_factory=validation,
         result_tracker=result_tracker,
         **stopper_kwargs,
     )
@@ -787,8 +1011,14 @@ def pipeline(  # noqa: C901
 
     # Add logging for debugging
     logging.debug("Run Pipeline based on following config:")
-    logging.debug(f"dataset: {dataset}")
-    logging.debug(f"dataset_kwargs: {dataset_kwargs}")
+    if dataset is not None:
+        logging.debug(f"dataset: {dataset}")
+        logging.debug(f"dataset_kwargs: {dataset_kwargs}")
+    else:
+        logging.debug('training: %s', training)
+        logging.debug('testing: %s', testing)
+        if validation:
+            logging.debug('validation: %s', validation)
     logging.debug(f"model: {model}")
     logging.debug(f"model_kwargs: {model_kwargs}")
     logging.debug(f"loss: {loss}")
@@ -817,9 +1047,9 @@ def pipeline(  # noqa: C901
     training_end_time = time.time() - training_start_time
 
     if use_testing_data:
-        mapped_triples = testing_triples_factory.mapped_triples
+        mapped_triples = testing.mapped_triples
     else:
-        mapped_triples = validation_triples_factory.mapped_triples
+        mapped_triples = validation.mapped_triples
 
     # Evaluate
     # Reuse optimal evaluation parameters from training if available
