@@ -4,22 +4,33 @@
 
 import gc
 import logging
+import pathlib
+import pickle
+import random
+import time
 from abc import ABC, abstractmethod
+from datetime import datetime
+from hashlib import md5
 from typing import Any, List, Mapping, Optional, Tuple, Type, Union
 
+import numpy as np
 import torch
 from torch.optim.optimizer import Optimizer
 from torch.utils.data import DataLoader
+from tqdm.autonotebook import tqdm, trange
 
-from ..losses import Loss
-from ..models.base import Model
+from ..constants import PYKEEN_CHECKPOINTS, PYKEEN_DEFAULT_CHECKPOINT
+from ..losses import Loss, has_mr_loss, has_nssa_loss
+from ..models import Model
 from ..stoppers import Stopper
-from ..tqdmw import tqdm, trange
 from ..trackers import ResultTracker
 from ..training.schlichtkrull_sampler import GraphSampler
 from ..triples import Instances, TriplesFactory
 from ..typing import MappedTriples
-from ..utils import is_cuda_oom_error, is_cudnn_error, normalize_string
+from ..utils import (
+    format_relative_comparison, get_batchnorm_modules, is_cuda_oom_error, is_cudnn_error,
+    normalize_string,
+)
 
 __all__ = [
     'TrainingLoop',
@@ -39,6 +50,10 @@ class TrainingApproachLossMismatchError(TypeError):
     """An exception when an illegal loss function is used with a given training approach."""
 
 
+class CheckpointMismatchError(RuntimeError):
+    """An exception when a provided checkpoint file does not match the current training loop setup."""
+
+
 class SubBatchingNotSupportedError(NotImplementedError):
     """An exception raised when sub batching is not implemented."""
 
@@ -47,8 +62,10 @@ class SubBatchingNotSupportedError(NotImplementedError):
         self.model = model
 
     def __str__(self):  # noqa: D105
-        return f'No sub-batching support for {self.model.__class__.__name__} due to modules ' \
-               f'{self.model.modules_not_supporting_sub_batching}.'
+        return (
+            f'No sub-batching support for {self.model.__class__.__name__} due to modules '
+            f'{get_batchnorm_modules(self.model)}.'
+        )
 
 
 def _get_optimizer_kwargs(optimizer: Optimizer) -> Mapping[str, Any]:
@@ -77,16 +94,21 @@ class TrainingLoop(ABC):
         self,
         model: Model,
         optimizer: Optional[Optimizer] = None,
+        automatic_memory_optimization: bool = True,
     ) -> None:
         """Initialize the training loop.
 
         :param model: The model to train
         :param optimizer: The optimizer to use while training the model
+        :param automatic_memory_optimization: bool
+            Whether to automatically optimize the sub-batch size during
+            training and batch size during evaluation with regards to the hardware at hand.
         """
         self.model = model
         self.optimizer = optimizer
         self.training_instances = None
         self.losses_per_epochs = []
+        self.automatic_memory_optimization = automatic_memory_optimization
 
         if self.loss_blacklist and isinstance(self.model.loss, tuple(self.loss_blacklist)):
             raise TrainingApproachLossMismatchError(
@@ -94,12 +116,16 @@ class TrainingLoop(ABC):
                 f' with training approach {self.__class__.__name__}',
             )
 
-        if self.model.is_mr_loss:
+        if has_mr_loss(self.model):
             self._loss_helper = self._mr_loss_helper
-        elif self.model.is_nssa_loss:
+        elif has_nssa_loss(self.model):
             self._loss_helper = self._self_adversarial_negative_sampling_loss_helper
         else:
-            self._loss_helper = self._label_loss_helper
+            self._loss_helper = self._label_loss_helper  # type: ignore
+
+        # The internal epoch state tracks the last finished epoch of the training loop to allow for
+        # seamless loading and saving of training checkpoints
+        self._epoch = 0
 
     @classmethod
     def get_normalized_name(cls) -> str:
@@ -116,6 +142,14 @@ class TrainingLoop(ABC):
         """The device used by the model."""
         return self.model.device
 
+    @property
+    def checksum(self) -> str:  # noqa: D401
+        """The checksum of the model and optimizer the training loop was configured with."""
+        h = md5()  # noqa: S303
+        h.update(str(self.model).encode('utf-8'))
+        h.update(str(self.optimizer).encode('utf-8'))
+        return h.hexdigest()
+
     def train(
         self,
         num_epochs: int = 1,
@@ -125,13 +159,20 @@ class TrainingLoop(ABC):
         sampler: Optional[str] = None,
         continue_training: bool = False,
         only_size_probing: bool = False,
+        use_tqdm: bool = True,
+        use_tqdm_batch: bool = True,
         tqdm_kwargs: Optional[Mapping[str, Any]] = None,
         stopper: Optional[Stopper] = None,
         result_tracker: Optional[ResultTracker] = None,
         sub_batch_size: Optional[int] = None,
         num_workers: Optional[int] = None,
         clear_optimizer: bool = False,
-    ) -> List[float]:
+        checkpoint_directory: Union[None, str, pathlib.Path] = None,
+        checkpoint_name: Optional[str] = None,
+        checkpoint_frequency: Optional[int] = None,
+        checkpoint_on_failure: bool = False,
+        drop_last: Optional[bool] = None,
+    ) -> Optional[List[float]]:
         """Train the KGE model.
 
         :param num_epochs:
@@ -150,6 +191,8 @@ class TrainingLoop(ABC):
             If set to False, (re-)initialize the model's weights. Otherwise continue training.
         :param only_size_probing:
             The evaluation is only performed for two batches to test the memory footprint, especially on GPUs.
+        :param use_tqdm: Should a progress bar be shown for epochs?
+        :param use_tqdm_batch: Should a progress bar be shown for batching (inside the epoch progress bar)?
         :param tqdm_kwargs:
             Keyword arguments passed to :mod:`tqdm` managing the progress bar.
         :param stopper:
@@ -164,9 +207,28 @@ class TrainingLoop(ABC):
         :param clear_optimizer:
             Whether to delete the optimizer instance after training (as the optimizer might have additional memory
             consumption due to e.g. moments in Adam).
+        :param checkpoint_directory:
+            An optional directory to store the checkpoint files. If None, a subdirectory named ``checkpoints`` in the
+            directory defined by :data:`pykeen.constants.PYKEEN_HOME` is used. Unless the environment variable
+            ``PYKEEN_HOME`` is overridden, this will be ``~/.pykeen/checkpoints``.
+        :param checkpoint_name:
+            The filename for saving checkpoints. If the given filename exists already, that file will be loaded and used
+            to continue training.
+        :param checkpoint_frequency:
+            The frequency of saving checkpoints in minutes. Setting it to 0 will save a checkpoint after every epoch.
+        :param checkpoint_on_failure:
+            Whether to save a checkpoint in cases of a RuntimeError or MemoryError. This option differs from ordinary
+            checkpoints, since ordinary checkpoints are only saved after a successful epoch. When saving checkpoints
+            due to failure of the training loop there is no guarantee that all random states can be recovered correctly,
+            which might cause problems with regards to the reproducibility of that specific training loop. Therefore,
+            these checkpoints are saved with a distinct checkpoint name, which will be
+            ``PyKEEN_just_saved_my_day_{datetime}.pt`` in the given checkpoint_root.
+        :param drop_last:
+            Whether to drop the last batch in each epoch to prevent smaller batches. Defaults to False, except if the
+            model contains batch normalization layers. Can be provided explicitly to override.
 
         :return:
-            A pair of the KGE model and the losses per epoch.
+            The losses per epoch.
         """
         # Create training instances
         # During size probing the training instances should not show the tqdm progress bar
@@ -175,20 +237,76 @@ class TrainingLoop(ABC):
         # In some cases, e.g. using Optuna for HPO, the cuda cache from a previous run is not cleared
         torch.cuda.empty_cache()
 
-        result = self._train(
-            num_epochs=num_epochs,
-            batch_size=batch_size,
-            slice_size=slice_size,
-            label_smoothing=label_smoothing,
-            sampler=sampler,
-            continue_training=continue_training,
-            only_size_probing=only_size_probing,
-            tqdm_kwargs=tqdm_kwargs,
-            stopper=stopper,
-            result_tracker=result_tracker,
-            sub_batch_size=sub_batch_size,
-            num_workers=num_workers,
-        )
+        # A checkpoint root is always created to ensure a fallback checkpoint can be saved
+        if checkpoint_directory is None:
+            checkpoint_directory = PYKEEN_CHECKPOINTS
+        checkpoint_directory = pathlib.Path(checkpoint_directory)
+        checkpoint_directory.mkdir(parents=True, exist_ok=True)
+        logger.debug('using checkpoint_root at %s', checkpoint_directory)
+
+        # If a checkpoint file is given, it must be loaded if it exists already
+        save_checkpoints = False
+        checkpoint_path = None
+        if checkpoint_name:
+            checkpoint_path = checkpoint_directory.joinpath(checkpoint_name)
+            if checkpoint_path.is_file():
+                self._load_state(path=checkpoint_path)
+                if stopper is not None:
+                    stopper_dict = stopper.load_summary_dict_from_training_loop_checkpoint(path=checkpoint_path)
+                    # If the stopper dict has any keys, those are written back to the stopper
+                    if stopper_dict:
+                        stopper._write_from_summary_dict(**stopper_dict)
+                    else:
+                        logger.warning(
+                            'the training loop was configured with a stopper but no stopper configuration was '
+                            'saved in the checkpoint',
+                        )
+                continue_training = True
+            else:
+                logger.info(f"=> no checkpoint found at '{checkpoint_path}'. Creating a new file.")
+            # The checkpoint frequency needs to be set to save checkpoints
+            if checkpoint_frequency is None:
+                checkpoint_frequency = 30
+            save_checkpoints = True
+        elif checkpoint_frequency is not None:
+            logger.warning(
+                "A checkpoint frequency was set, but no checkpoint file was given. No checkpoints will be created",
+            )
+
+        checkpoint_on_failure_file_path = None
+        if checkpoint_on_failure:
+            # In case a checkpoint frequency was set, we warn that no checkpoints will be saved
+            date_string = datetime.now().strftime('%Y%m%d_%H_%M_%S')
+            # If no checkpoints were requested, a fallback checkpoint is set in case the training loop crashes
+            checkpoint_on_failure_file_path = checkpoint_directory.joinpath(
+                PYKEEN_DEFAULT_CHECKPOINT.replace('.', f"_{date_string}."),
+            )
+
+        # If the stopper loaded from the training loop checkpoint stopped the training, we return those results
+        if getattr(stopper, 'stopped', False):
+            result: Optional[List[float]] = self.losses_per_epochs
+        else:
+            result = self._train(
+                num_epochs=num_epochs,
+                batch_size=batch_size,
+                slice_size=slice_size,
+                label_smoothing=label_smoothing,
+                sampler=sampler,
+                continue_training=continue_training,
+                only_size_probing=only_size_probing,
+                use_tqdm=use_tqdm,
+                use_tqdm_batch=use_tqdm_batch,
+                tqdm_kwargs=tqdm_kwargs,
+                stopper=stopper,
+                result_tracker=result_tracker,
+                sub_batch_size=sub_batch_size,
+                num_workers=num_workers,
+                save_checkpoints=save_checkpoints,
+                checkpoint_path=checkpoint_path,
+                checkpoint_frequency=checkpoint_frequency,
+                checkpoint_on_failure_file_path=checkpoint_on_failure_file_path,
+                drop_last=drop_last,
+            )
 
         # Ensure the release of memory
         torch.cuda.empty_cache()
@@ -208,12 +326,19 @@ class TrainingLoop(ABC):
         sampler: Optional[str] = None,
         continue_training: bool = False,
         only_size_probing: bool = False,
+        use_tqdm: bool = True,
+        use_tqdm_batch: bool = True,
         tqdm_kwargs: Optional[Mapping[str, Any]] = None,
         stopper: Optional[Stopper] = None,
         result_tracker: Optional[ResultTracker] = None,
         sub_batch_size: Optional[int] = None,
         num_workers: Optional[int] = None,
-    ) -> List[float]:
+        save_checkpoints: bool = False,
+        checkpoint_path: Union[None, str, pathlib.Path] = None,
+        checkpoint_frequency: Optional[int] = None,
+        checkpoint_on_failure_file_path: Union[None, str, pathlib.Path] = None,
+        drop_last: Optional[bool] = None,
+    ) -> Optional[List[float]]:
         """Train the KGE model.
 
         :param num_epochs:
@@ -232,6 +357,10 @@ class TrainingLoop(ABC):
             If set to False, (re-)initialize the model's weights. Otherwise continue training.
         :param only_size_probing:
             The evaluation is only performed for two batches to test the memory footprint, especially on GPUs.
+        :param use_tqdm:
+            Turn on the progress bar for epochs
+        :param use_tqdm_batch:
+            Turn on the progress bar for batches (sub-progress bar for epochs)
         :param tqdm_kwargs:
             Keyword arguments passed to :mod:`tqdm` managing the progress bar.
         :param stopper:
@@ -243,20 +372,51 @@ class TrainingLoop(ABC):
             If provided split each batch into sub-batches to avoid memory issues for large models / small GPUs.
         :param num_workers:
             The number of child CPU workers used for loading data. If None, data are loaded in the main process.
+        :param save_checkpoints:
+            Activate saving checkpoints.
+        :param checkpoint_path:
+            The full filepath for saving checkpoints.
+        :param checkpoint_frequency:
+            The frequency of saving checkpoints in minutes. Setting it to 0 will save a checkpoint after every epoch.
+        :param checkpoint_on_failure_file_path:
+            The full filepath for saving checkpoints on failure.
+        :param drop_last:
+            Whether to drop the last batch in each epoch to prevent smaller batches. Defaults to False, except if the
+            model contains batch normalization layers. Can be provided explicitly to override.
 
         :return:
-            A pair of the KGE model and the losses per epoch.
+            The losses per epoch.
         """
+        if self.training_instances is None:
+            raise ValueError('must set training instances before running _train()')
+        if self.optimizer is None:
+            raise ValueError('optimizer must be set before running _train()')
+
         # Take the biggest possible training batch_size, if batch_size not set
         batch_size_sufficient = False
         if batch_size is None:
-            if self.model.automatic_memory_optimization:
-                batch_size, batch_size_sufficient = self.batch_size_search()
+            if self.automatic_memory_optimization:
+                # Using automatic memory optimization on CPU may result in undocumented crashes due to OS' OOM killer.
+                if self.model.device.type == 'cpu':
+                    batch_size = 256
+                    batch_size_sufficient = True
+                    logger.info(
+                        "Currently automatic memory optimization only supports GPUs, but you're using a CPU. "
+                        "Therefore, the batch_size will be set to the default value '{batch_size}'",
+                    )
+                else:
+                    batch_size, batch_size_sufficient = self.batch_size_search()
             else:
                 batch_size = 256
+                logger.info(f"No batch_size provided. Setting batch_size to '{batch_size}'.")
 
         # This will find necessary parameters to optimize the use of the hardware at hand
-        if not only_size_probing and self.model.automatic_memory_optimization and not batch_size_sufficient:
+        if (
+            not only_size_probing
+            and self.automatic_memory_optimization
+            and not batch_size_sufficient
+            and not continue_training
+        ):
             # return the relevant parameters slice_size and batch_size
             sub_batch_size, slice_size = self.sub_batch_and_slice(batch_size)
 
@@ -266,11 +426,22 @@ class TrainingLoop(ABC):
 
         if sub_batch_size is None or sub_batch_size == batch_size:  # by default do not split batches in sub-batches
             sub_batch_size = batch_size
-        elif not self.model.supports_subbatching:
+        elif get_batchnorm_modules(self.model):  # if there are any, this is truthy
             raise SubBatchingNotSupportedError(self.model)
 
+        model_contains_batch_norm = bool(get_batchnorm_modules(self.model))
+        if batch_size == 1 and model_contains_batch_norm:
+            raise ValueError("Cannot train a model with batch_size=1 containing BatchNorm layers.")
+        if drop_last is None:
+            drop_last = model_contains_batch_norm
+            if drop_last and not only_size_probing:
+                logger.info(
+                    "Dropping last (incomplete) batch each epoch (%s batches).",
+                    format_relative_comparison(part=1, total=len(self.training_instances)),
+                )
+
         # Sanity check
-        if self.model.is_mr_loss and label_smoothing > 0.:
+        if has_mr_loss(self.model) and label_smoothing > 0.:
             raise RuntimeError('Label smoothing can not be used with margin ranking loss.')
 
         # Force weight initialization if training continuation is not explicitly requested.
@@ -282,13 +453,13 @@ class TrainingLoop(ABC):
             optimizer_kwargs = _get_optimizer_kwargs(self.optimizer)
             self.optimizer = self.optimizer.__class__(
                 params=self.model.get_grad_params(),
-                **optimizer_kwargs
+                **optimizer_kwargs,
             )
         elif not self.optimizer.state:
             raise ValueError('Cannot continue_training without being trained once.')
 
         # Ensure the model is on the correct device
-        self.model: Model = self.model.to(self.device)
+        self.model = self.model.to(self.device)
 
         # Create Sampler
         if sampler == 'schlichtkrull':
@@ -302,19 +473,24 @@ class TrainingLoop(ABC):
             num_workers = 0
 
         # Bind
-        num_training_instances = self.training_instances.num_instances
+        num_training_instances = len(self.training_instances)
+
+        _use_outer_tqdm = not only_size_probing and use_tqdm
+        _use_inner_tqdm = _use_outer_tqdm and use_tqdm_batch
 
         # When size probing, we don't want progress bars
-        if not only_size_probing:
+        if _use_outer_tqdm:
             # Create progress bar
             _tqdm_kwargs = dict(desc=f'Training epochs on {self.device}', unit='epoch')
             if tqdm_kwargs is not None:
                 _tqdm_kwargs.update(tqdm_kwargs)
-            epochs = trange(1, 1 + num_epochs, **_tqdm_kwargs)
-            logger.info(f'using stopper: {stopper}')
-        else:
+            epochs = trange(self._epoch + 1, 1 + num_epochs, **_tqdm_kwargs, initial=self._epoch, total=num_epochs)
+        elif only_size_probing:
             epochs = range(1, 1 + num_epochs)
-            logger.debug(f'using stopper: {stopper}')
+        else:
+            epochs = range(self._epoch + 1, 1 + num_epochs)
+
+        logger.debug(f'using stopper: {stopper}')
 
         train_data_loader = DataLoader(
             sampler=sampler,
@@ -322,85 +498,128 @@ class TrainingLoop(ABC):
             batch_size=batch_size,
             shuffle=shuffle,
             num_workers=num_workers,
+            drop_last=drop_last,
         )
 
+        # Save the time to track when the saved point was available
+        last_checkpoint = time.time()
+
         # Training Loop
         for epoch in epochs:
-            # Enforce training mode
-            self.model.train()
+            # When training with an early stopper the memory pressure changes, which may allow for errors each epoch
+            try:
+                # Enforce training mode
+                self.model.train()
 
-            # Accumulate loss over epoch
-            current_epoch_loss = 0.
+                # Accumulate loss over epoch
+                current_epoch_loss = 0.
 
-            # Batching
-            # Only create a progress bar when not in size probing mode
-            if not only_size_probing:
-                batches = tqdm(train_data_loader, desc=f'Training batches on {self.device}', leave=False, unit='batch')
-            else:
-                batches = train_data_loader
-
-            # Flag to check when to quit the size probing
-            evaluated_once = False
-
-            for batch in batches:
-                # Recall that torch *accumulates* gradients. Before passing in a
-                # new instance, you need to zero out the gradients from the old instance
-                self.optimizer.zero_grad()
-
-                # Get batch size of current batch (last batch may be incomplete)
-                current_batch_size = self._get_batch_size(batch)
-
-                # accumulate gradients for whole batch
-                for start in range(0, current_batch_size, sub_batch_size):
-                    stop = min(start + sub_batch_size, current_batch_size)
-
-                    # forward pass call
-                    current_epoch_loss += self._forward_pass(
-                        batch,
-                        start,
-                        stop,
-                        current_batch_size,
-                        label_smoothing,
-                        slice_size,
+                # Batching
+                # Only create a progress bar when not in size probing mode
+                if _use_inner_tqdm:
+                    batches = tqdm(
+                        train_data_loader,
+                        desc=f'Training batches on {self.device}',
+                        leave=False,
+                        unit='batch',
                     )
+                else:
+                    batches = train_data_loader
 
-                # when called by batch_size_search(), the parameter update should not be applied.
-                if not only_size_probing:
-                    # update parameters according to optimizer
-                    self.optimizer.step()
+                # Flag to check when to quit the size probing
+                evaluated_once = False
 
-                # After changing applying the gradients to the embeddings, the model is notified that the forward
-                # constraints are no longer applied
-                self.model.post_parameter_update()
+                for batch in batches:
+                    # Recall that torch *accumulates* gradients. Before passing in a
+                    # new instance, you need to zero out the gradients from the old instance
+                    self.optimizer.zero_grad()
 
-                # For testing purposes we're only interested in processing one batch
-                if only_size_probing and evaluated_once:
-                    break
+                    # Get batch size of current batch (last batch may be incomplete)
+                    current_batch_size = self._get_batch_size(batch)
 
-                evaluated_once = True
+                    # accumulate gradients for whole batch
+                    for start in range(0, current_batch_size, sub_batch_size):
+                        stop = min(start + sub_batch_size, current_batch_size)
 
-            del batch
-            del batches
-            gc.collect()
-            self.optimizer.zero_grad()
-            self._free_graph_and_cache()
+                        # forward pass call
+                        current_epoch_loss += self._forward_pass(
+                            batch,
+                            start,
+                            stop,
+                            current_batch_size,
+                            label_smoothing,
+                            slice_size,
+                        )
 
-            # When size probing we don't need the losses
-            if only_size_probing:
-                return None
+                    # when called by batch_size_search(), the parameter update should not be applied.
+                    if not only_size_probing:
+                        # update parameters according to optimizer
+                        self.optimizer.step()
 
-            # Track epoch loss
-            epoch_loss = current_epoch_loss / num_training_instances
-            self.losses_per_epochs.append(epoch_loss)
-            result_tracker.log_metrics({'loss': epoch_loss}, step=epoch)
+                    # After changing applying the gradients to the embeddings, the model is notified that the forward
+                    # constraints are no longer applied
+                    self.model.post_parameter_update()
 
-            # Print loss information to console
-            epochs.set_postfix({
-                'loss': self.losses_per_epochs[-1],
-                'prev_loss': self.losses_per_epochs[-2] if epoch > 2 else float('nan'),
-            })
+                    # For testing purposes we're only interested in processing one batch
+                    if only_size_probing and evaluated_once:
+                        break
 
-            if stopper is not None and stopper.should_evaluate(epoch) and stopper.should_stop():
+                    evaluated_once = True
+
+                del batch
+                del batches
+                gc.collect()
+                self.optimizer.zero_grad()
+                self._free_graph_and_cache()
+
+                # When size probing we don't need the losses
+                if only_size_probing:
+                    return None
+
+                # Track epoch loss
+                epoch_loss = current_epoch_loss / num_training_instances
+                self.losses_per_epochs.append(epoch_loss)
+                result_tracker.log_metrics({'loss': epoch_loss}, step=epoch)
+
+                # Print loss information to console
+                if _use_outer_tqdm:
+                    epochs.set_postfix({
+                        'loss': self.losses_per_epochs[-1],
+                        'prev_loss': self.losses_per_epochs[-2] if epoch > 2 else float('nan'),
+                    })
+
+                # Save the last successful finished epoch
+                self._epoch = epoch
+
+                should_stop = False
+                if stopper is not None and stopper.should_evaluate(epoch) and stopper.should_stop(epoch):
+                    should_stop = True
+            # When the training loop failed, a fallback checkpoint is created to resume training.
+            except (MemoryError, RuntimeError) as e:
+                logger.warning(f'The training loop just failed during epoch {epoch} due to error {str(e)}.')
+                if checkpoint_on_failure_file_path:
+                    self._save_state(path=checkpoint_on_failure_file_path, stopper=stopper)
+                    logger.warning(
+                        "However, don't worry we got you covered. PyKEEN just saved a checkpoint when this happened "
+                        f"at '{checkpoint_on_failure_file_path}'. To resume training from the checkpoint file just "
+                        f"restart your code and pass this file path to the training loop or pipeline you used "
+                        f"as 'checkpoint_file' argument.",
+                    )
+                raise e
+
+            # If a checkpoint file is given, we check whether it is time to save a checkpoint
+            if save_checkpoints:
+                minutes_since_last_checkpoint = (time.time() - last_checkpoint) // 60
+                # MyPy overrides are because you should
+                if (
+                    minutes_since_last_checkpoint >= checkpoint_frequency  # type: ignore
+                    or should_stop
+                    or epoch == num_epochs
+                ):
+                    self._save_state(path=checkpoint_path, stopper=stopper)  # type: ignore
+                    last_checkpoint = time.time()
+
+            if should_stop:
                 return self.losses_per_epochs
 
         return self.losses_per_epochs
@@ -412,7 +631,7 @@ class TrainingLoop(ABC):
             start=start,
             stop=stop,
             label_smoothing=label_smoothing,
-            slice_size=slice_size
+            slice_size=slice_size,
         )
 
         # raise error when non-finite loss occurs (NaN, +/-inf)
@@ -428,8 +647,8 @@ class TrainingLoop(ABC):
         loss.backward()
         current_epoch_loss = loss.item()
 
-        # reset the regularizer to free the computational graph
-        self.model.regularizer.reset()
+        self.model.post_forward_pass()
+        # TODO why not call torch.cuda.empty_cache()? or call self._free_graph_and_cache()?
 
         return current_epoch_loss
 
@@ -515,19 +734,18 @@ class TrainingLoop(ABC):
 
         return batch_size, evaluated_once
 
-    def sub_batch_and_slice(self, batch_size: int) -> Tuple[int, int]:
+    def sub_batch_and_slice(self, batch_size: int) -> Tuple[int, Optional[int]]:
         """Check if sub-batching and/or slicing is necessary to train the model on the hardware at hand."""
         sub_batch_size, finished_search, supports_sub_batching = self._sub_batch_size_search(batch_size=batch_size)
         # If the sub_batch_size did not finish search with a possibility that fits the hardware, we have to try slicing
-        if not finished_search:
-            slice_size = self._slice_size_search(
-                batch_size=batch_size,
-                sub_batch_size=sub_batch_size,
-                supports_sub_batching=supports_sub_batching,
-            )
-        else:
-            slice_size = None
+        if finished_search:
+            return sub_batch_size, None
 
+        slice_size = self._slice_size_search(
+            batch_size=batch_size,
+            sub_batch_size=sub_batch_size,
+            supports_sub_batching=supports_sub_batching,
+        )
         return sub_batch_size, slice_size
 
     @abstractmethod
@@ -588,7 +806,7 @@ class TrainingLoop(ABC):
 
         if not finished_search:
             logger.info('Starting sub_batch_size search for training now...')
-            if not self.model.supports_subbatching:
+            if get_batchnorm_modules(self.model):  # if there are any, this is truthy
                 logger.info('This model does not support sub-batching.')
                 supports_sub_batching = False
                 sub_batch_size = batch_size
@@ -601,7 +819,7 @@ class TrainingLoop(ABC):
                             num_epochs=1,
                             batch_size=batch_size,
                             sub_batch_size=sub_batch_size,
-                            only_size_probing=True
+                            only_size_probing=True,
                         )
                     except RuntimeError as runtime_error:
                         self._free_graph_and_cache()
@@ -609,7 +827,8 @@ class TrainingLoop(ABC):
                             raise runtime_error
                         if sub_batch_size == 1:
                             logger.info(
-                                f"Even sub_batch_size={sub_batch_size} does not fit in memory with these parameters")
+                                f"Even sub_batch_size={sub_batch_size} does not fit in memory with these parameters",
+                            )
                             break
                         logger.debug(f'The sub_batch_size {sub_batch_size} was too big, trying less now.')
                         sub_batch_size //= 2
@@ -646,17 +865,96 @@ class TrainingLoop(ABC):
     ) -> torch.FloatTensor:
         raise NotImplementedError
 
-    def to_embeddingdb(self, session=None, use_tqdm: bool = False):
-        """Upload to the embedding database.
-
-        :param session: Optional SQLAlchemy session
-        :param use_tqdm: Use :mod:`tqdm` progress bar?
-        :rtype: embeddingdb.sql.models.Collection
-        """
-        return self.model.to_embeddingdb(session=session, use_tqdm=use_tqdm)
-
     def _free_graph_and_cache(self):
-        # The regularizer has to be reset to free the computational graph
-        self.model.regularizer.reset()
+        self.model._free_graph_and_cache()
         # The cache of the previous run has to be freed to allow accurate memory availability estimates
         torch.cuda.empty_cache()
+
+    def _save_state(self, path: Union[str, pathlib.Path], stopper: Optional[Stopper] = None) -> None:
+        """Save the state of the training loop.
+
+        :param path:
+            Path of the file where to store the state in.
+        :param stopper:
+            An instance of :class:`pykeen.stopper.EarlyStopper` with settings for checking
+            if training should stop early
+        """
+        if self.optimizer is None:
+            raise ValueError
+
+        logger.debug("=> Saving checkpoint.")
+
+        if stopper is None:
+            stopper_dict: Mapping[str, Any] = dict()
+        else:
+            stopper_dict = stopper.get_summary_dict()
+
+        # Only if a cuda device is available, the random state is accessed
+        if torch.cuda.is_available():
+            torch_cuda_random_state = torch.cuda.get_rng_state()
+        else:
+            torch_cuda_random_state = None
+
+        torch.save(
+            {
+                'epoch': self._epoch,
+                'loss': self.losses_per_epochs,
+                'model_state_dict': self.model.state_dict(),
+                'optimizer_state_dict': self.optimizer.state_dict(),
+                'checksum': self.checksum,
+                'random_seed': self.model._random_seed,
+                'stopper_dict': stopper_dict,
+                'random_state': random.getstate(),
+                'np_random_state': np.random.get_state(),
+                'torch_random_state': torch.random.get_rng_state(),
+                'torch_cuda_random_state': torch_cuda_random_state,
+            },
+            path,
+            pickle_protocol=pickle.HIGHEST_PROTOCOL,
+        )
+        logger.info(f"=> Saved checkpoint after having finished epoch {self._epoch}.")
+
+    def _load_state(self, path: Union[str, pathlib.Path]) -> None:
+        """Load the state of the training loop from a checkpoint.
+
+        :param path:
+            Path of the file where to load the state from.
+
+        :raises CheckpointMismatchError:
+            If the given checkpoint file has a non-matching checksum, i.e. it was saved with a different configuration.
+        """
+        if self.optimizer is None:
+            raise ValueError
+
+        logger.info(f"=> loading checkpoint '{path}'")
+        checkpoint = torch.load(path)
+        if checkpoint['checksum'] != self.checksum:
+            raise CheckpointMismatchError(
+                f"The checkpoint file '{path}' that was provided already exists, but seems to be "
+                f"from a different training loop setup.",
+            )
+        # Cuda requires its own random state, which can only be set when a cuda device is available
+        torch_cuda_random_state = checkpoint['torch_cuda_random_state']
+        if torch_cuda_random_state is not None and torch.cuda.is_available():
+            torch.cuda.set_rng_state(torch_cuda_random_state)
+        elif torch_cuda_random_state is not None and not torch.cuda.is_available():
+            logger.warning(
+                "You're currently trying to resume the training loop on a CPU from a checkpoint that was saved "
+                "with a GPU. Therefore, the random state for the CUDA devices can't be set and results may not "
+                "be deterministic.",
+            )
+        elif torch_cuda_random_state is None and torch.cuda.is_available():
+            logger.warning(
+                "You're currently trying to resume the training loop on a GPU from a checkpoint that was saved "
+                "without a GPU. Therefore, the random state for the CUDA devices won't be set and results may not "
+                "be deterministic.",
+            )
+
+        self._epoch = checkpoint['epoch']
+        self.losses_per_epochs = checkpoint['loss']
+        self.model.load_state_dict(checkpoint['model_state_dict'])
+        self.optimizer.load_state_dict(checkpoint['optimizer_state_dict'])
+        random.setstate(checkpoint['random_state'])
+        np.random.set_state(checkpoint['np_random_state'])
+        torch.random.set_rng_state(checkpoint['torch_random_state'])
+        logger.info(f"=> loaded checkpoint '{path}' stopped after having finished epoch {checkpoint['epoch']}")
